# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math
import warnings

from torch.optim.optimizer import Optimizer

from .torch_backports import OneCycleLR


class MesaOneCycleLR(OneCycleLR):
    r"""
    A 1-cycle learning rate schedule with a flat region at maximum learning rate.

    Sets the learning rate of each parameter group according to the
    1cycle learning rate policy. The 1cycle policy anneals the learning
    rate from an initial learning rate to some maximum learning rate and then
    from that maximum learning rate to some minimum learning rate much lower
    than the initial learning rate.
    This policy was initially described in the paper `Super-Convergence:
    Very Fast Training of Neural Networks Using Large Learning Rates`_.

    The 1cycle learning rate policy changes the learning rate after every batch.
    ``step`` should be called after a batch has been used for training.

    This scheduler is not chainable.

    Note also that the total number of steps in the cycle can be determined in one
    of two ways (listed in order of precedence):

    #. A value for total_steps is explicitly provided.
    #. A number of epochs (epochs) and a number of steps per epoch
       (steps_per_epoch) are provided.
       In this case, the number of total steps is inferred by
       total_steps = epochs * steps_per_epoch

    You must either provide a value for total_steps or provide a value for both
    epochs and steps_per_epoch.

    Parameters
    ----------
    optimizer : :class:`torch.optim.optimizer.Optimizer`
        Wrapped optimizer.
    max_lr : float or list
        Upper learning rate boundaries in the cycle
        for each parameter group.
    total_steps : int
        The total number of steps in the cycle. Note that
        if a value is provided here, then it must be inferred by providing
        a value for epochs and steps_per_epoch.
        Default: None
    epochs : int
        The number of epochs to train for. This is used along
        with steps_per_epoch in order to infer the total number of steps in the cycle
        if a value for total_steps is not provided.
        Default: None
    steps_per_epoch : int
        The number of steps per epoch to train for. This is
        used along with epochs in order to infer the total number of steps in the
        cycle if a value for total_steps is not provided.
        Default: None
    pct_start : float
        The percentage of the cycle (in number of steps) spent
        increasing the learning rate.
        Default: 0.25
    pct_end : float
        The percentage of the cycle (in number of steps) spent
        before decreasing the learning rate.
        Default: 0.75
    anneal_strategy : {"cos", "linear"}
        Specifies the annealing strategy: "cos" for cosine annealing, "linear" for
        linear annealing.
        Default: "cos".
    cycle_momentum : bool, default=True
        If ``True``, momentum is cycled inversely
        to learning rate between "base_momentum" and "max_momentum".
        Default: True
    base_momentum : float or list
        Lower momentum boundaries in the cycle
        for each parameter group. Note that momentum is cycled inversely
        to learning rate; at the peak of a cycle, momentum is
        "base_momentum" and learning rate is "max_lr".
        Default: 0.85
    max_momentum : float or list
        Upper momentum boundaries in the cycle
        for each parameter group. Functionally,
        it defines the cycle amplitude (max_momentum - base_momentum).
        Note that momentum is cycled inversely
        to learning rate; at the start of a cycle, momentum is "max_momentum"
        and learning rate is "base_lr"
        Default: 0.95
    div_factor : float
        Determines the initial learning rate via
        initial_lr = max_lr/div_factor
        Default: 25
    final_div_factor : float
        Determines the minimum learning rate via
        min_lr = initial_lr/final_div_factor
        Default: 1e4
    last_epoch : int
        The index of the last batch. This parameter is used when
        resuming a training job. Since ``step()`` should be invoked after each
        batch instead of after each epoch, this number represents the total
        number of *batches* computed, not the total number of epochs computed.
        When last_epoch=-1, the schedule is started from the beginning.
        Default: -1

    Example
    -------
    >>> data_loader = torch.utils.data.DataLoader(...)
    >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)
    >>> scheduler = MesaOneCycleLR(optimizer, max_lr=0.01, steps_per_epoch=len(data_loader), epochs=10)
    >>> for epoch in range(10):
    >>>     for batch in data_loader:
    >>>         train_batch(...)
    >>>         scheduler.step()

    .. _Super-Convergence\: Very Fast Training of Neural Networks Using Large Learning Rates:
        https://arxiv.org/abs/1708.07120
    """

    def __init__(
        self,
        optimizer,
        max_lr,
        total_steps=None,
        pct_start=0.25,
        pct_end=0.75,
        **kwargs,
    ):

        # Validate pct_start
        if not isinstance(pct_start, float) or pct_start < 0 or pct_start > 1:
            raise ValueError(
                "Expected pct_start to be a float between 0 and 1, but got {}".format(
                    pct_start
                )
            )

        # Validate pct_end
        if pct_end < pct_start or pct_end > 1 or not isinstance(pct_start, float):
            raise ValueError(
                "Expected pct_end to be a float between pct_start={} and 1, but got {}".format(
                    pct_start, pct_end
                )
            )

        super(MesaOneCycleLR, self).__init__(
            optimizer, max_lr, total_steps=total_steps, pct_start=pct_start, **kwargs
        )

        self.step_size_up = float(pct_start * self.total_steps) - 1
        self.step_size_down_start = float(pct_end * self.total_steps) - 1
        self.step_size_down = float(self.total_steps - self.step_size_down_start) - 1

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, "
                "please use ``get_last_lr()``.",
                DeprecationWarning,
            )

        lrs = []
        step_num = self.last_epoch

        if step_num > self.total_steps:
            raise ValueError(
                "Tried to step {} times. The specified number of total steps is {}".format(
                    step_num + 1, self.total_steps
                )
            )

        for group in self.optimizer.param_groups:
            if step_num <= self.step_size_up:
                computed_lr = self.anneal_func(
                    group["initial_lr"], group["max_lr"], step_num / self.step_size_up
                )
                if self.cycle_momentum:
                    computed_momentum = self.anneal_func(
                        group["max_momentum"],
                        group["base_momentum"],
                        step_num / self.step_size_up,
                    )
            elif step_num <= self.step_size_down_start:
                computed_lr = group["max_lr"]
                if self.cycle_momentum:
                    computed_momentum = group["base_momentum"]
            else:
                down_step_num = step_num - self.step_size_down_start
                computed_lr = self.anneal_func(
                    group["max_lr"],
                    group["min_lr"],
                    down_step_num / self.step_size_down,
                )
                if self.cycle_momentum:
                    computed_momentum = self.anneal_func(
                        group["base_momentum"],
                        group["max_momentum"],
                        down_step_num / self.step_size_down,
                    )

            lrs.append(computed_lr)
            if self.cycle_momentum:
                if self.use_beta1:
                    _, beta2 = group["betas"]
                    group["betas"] = (computed_momentum, beta2)
                else:
                    group["momentum"] = computed_momentum

        return lrs



name = "echofilter"
path = name.lower().replace("-", "_").replace(" ", "_")
version = "1.1.0"
author = "Scott C. Lowe"
author_email = "scott.code.lowe@gmail.com"
description = "Remove echosounder noise by identifying the ocean floor and entrained air at the ocean surface."
url = "https://github.com/DeepSenseCA/echofilter"
license = "AGPLv3"



"""
Input/Output handling for raw Echoview files.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import csv
import datetime
import os
import textwrap
import warnings
from collections import OrderedDict

import numpy as np
import pandas as pd
import scipy.interpolate
import scipy.ndimage
import skimage.measure

from ..ui import style
from . import utils

ROOT_DATA_DIR = "/data/dsforce/surveyExports"

TRANSECT_FIELD_TYPES = {
    "Ping_index": int,
    "Distance_gps": float,
    "Distance_vl": float,
    "Ping_date": str,
    "Ping_time": str,
    "Ping_milliseconds": float,
    "Latitude": float,
    "Longitude": float,
    "Depth_start": float,
    "Depth_stop": float,
    "Range_start": float,
    "Range_stop": float,
    "Sample_count": int,
}


def transect_reader(fname):
    """
    Create a generator which iterates through a survey csv file.

    Parameters
    ----------
    fname: str
        Path to survey CSV file.

    Returns
    -------
    generator
        Yields a tupule of `(metadata, data)`, where metadata is a dict,
        and data is a :class:`numpy.ndarray`. Each yield corresponds to a single
        row in the data. Every row (except for the header) is yielded.
    """
    metadata_header = []
    with open(fname, "rb") as hf:
        for i_row, row in enumerate(hf):
            try:
                row = row.decode("utf-8-sig" if i_row == 0 else "utf-8")
            except Exception:
                if i_row == 0:
                    raise
                print(
                    "Row {} of {} contained a byte which is not in UTF-8"
                    " and will be skipped.".format(i_row, fname)
                )
                continue
            row = row.split(",")
            row = [entry.strip() for entry in row]
            if i_row == 0:
                metadata_header = row
                continue
            metadata = row[: len(metadata_header)]
            metadata_d = OrderedDict()
            for k, v in zip(metadata_header, metadata):
                if k in TRANSECT_FIELD_TYPES:
                    metadata_d[k] = TRANSECT_FIELD_TYPES[k](v)
                else:
                    metadata_d[k] = v
            data = np.array([float(x) for x in row[len(metadata_header) :]])
            yield metadata_d, data


def count_lines(filename):
    """
    Count the number of lines in a file.

    Parameters
    ----------
    filename : str
        Path to file.

    Returns
    -------
    int
        Number of lines in file.
    """
    with open(filename, "rb") as f:
        for _i, _ in enumerate(f):
            pass
    return _i + 1


def transect_loader(
    fname,
    skip_lines=0,
    warn_row_overflow=None,
    row_len_selector="mode",
):
    """
    Load an entire survey transect CSV.

    Parameters
    ----------
    fname : str
        Path to survey CSV file.
    skip_lines : int, optional
        Number of initial entries to skip. Default is 0.
    warn_row_overflow : bool or int, optional
        Whether to print a warning message if the number of elements in a
        row exceeds the expected number. If this is an int, this is the number
        of times to display the warnings before they are supressed. If this
        is ``True``, the number of outputs is unlimited. If ``None``, the
        maximum number of underflow and overflow warnings differ: if
        ``row_len_selector`` is ``"init"`` or ``"min"``, underflow always produces a
        message and the overflow messages stop at 2; otherwise the values are
        reversed. Default is ``None``.
    row_len_selector : {"init", "min", "max", "median", "mode"}, optional
        The method used to determine which row length (number of depth samples)
        to use. Default is ``"mode"``, the most common row length across all
        the measurement timepoints.

    Returns
    -------
    numpy.ndarray
        Timestamps for each row, in seconds. Note: not corrected for timezone
        (so make sure your timezones are internally consistent).
    numpy.ndarray
        Depth of each column, in metres.
    numpy.ndarray
        Survey signal (Sv, for instance). Units match that of the file.
    """
    row_len_selector = row_len_selector.lower()
    if row_len_selector in {"init", "min"}:
        expand_for_overflow = False
    else:
        expand_for_overflow = True

    if warn_row_overflow is True:
        warn_row_overflow = np.inf

    if warn_row_overflow is not None:
        warn_row_underflow = warn_row_overflow
    elif expand_for_overflow:
        warn_row_underflow = 2
        warn_row_overflow = np.inf
    else:
        warn_row_underflow = np.inf
        warn_row_overflow = 2

    # We remove one from the line count because of the header
    # which is excluded from output
    n_lines = count_lines(fname) - 1

    # Initialise output array
    for i_line, (_, row) in enumerate(transect_reader(fname)):
        if i_line < min(n_lines, max(1, skip_lines)):
            continue
        n_depths_init = len(row)
        break

    n_depth_exp = n_depths_init

    data = np.empty((n_lines - skip_lines, n_depth_exp))
    data[:] = np.nan
    timestamps = np.empty((n_lines - skip_lines))
    timestamps[:] = np.nan

    row_lengths = np.empty((n_lines - skip_lines), dtype=np.int)
    row_depth_starts = np.empty((n_lines - skip_lines))
    row_depth_ends = np.empty((n_lines - skip_lines))

    n_warn_overflow = 0
    n_warn_underflow = 0

    n_entry = 0
    for i_line, (meta, row) in enumerate(transect_reader(fname)):
        if i_line < skip_lines:
            continue
        i_entry = i_line - skip_lines

        # Track the range of depths used in the row with this length
        row_lengths[i_entry] = len(row)
        row_depth_starts[i_entry] = meta["Depth_start"]
        row_depth_ends[i_entry] = meta["Depth_stop"]

        if len(row) > n_depth_exp:
            if n_warn_overflow < warn_row_overflow:
                print(
                    "Row {} of {} exceeds expected n_depth of {} with {}".format(
                        i_line, fname, n_depth_exp, len(row)
                    )
                )
                n_warn_overflow += 1
            if expand_for_overflow:
                data = np.pad(
                    data,
                    ((0, 0), (0, len(row) - n_depth_exp)),
                    mode="constant",
                    constant_values=np.nan,
                )
                n_depth_exp = len(row)

        if len(row) < n_depth_exp:
            if n_warn_underflow < warn_row_underflow:
                print(
                    "Row {} of {} shorter than expected n_depth_exp of {} with {}".format(
                        i_line, fname, n_depth_exp, len(row)
                    )
                )
                n_warn_underflow += 1
            data[i_entry, : len(row)] = row
        else:
            data[i_entry, :] = row[:n_depth_exp]

        timestamps[i_entry] = datetime.datetime.strptime(
            "{}T{}.{:06d}".format(
                meta["Ping_date"],
                meta["Ping_time"],
                int(1000 * float(meta["Ping_milliseconds"])),
            ),
            "%Y-%m-%dT%H:%M:%S.%f",
        ).timestamp()
        n_entry += 1

    # Turn NaNs into NaNs (instead of extremely negative number)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "invalid value encountered in less")
        warnings.filterwarnings("ignore", "invalid value encountered in greater")
        # 9.9e+37 and -9.9e+37 are special values indicating missing data
        # https://support.echoview.com/WebHelp/Reference/File_formats/Export_file_formats/Special_Export_Values.htm
        data[data < -1e37] = np.nan
        data[data > 1e37] = np.nan

    # Trim timestamps dimension down to size
    timestamps = timestamps[:n_entry]
    data = data[:n_entry]
    row_lengths = row_lengths[:n_entry]
    row_depth_starts = row_depth_starts[:n_entry]
    row_depth_ends = row_depth_ends[:n_entry]

    # Work out what row length we should return
    if row_len_selector == "init":
        n_depth_use = n_depths_init
    elif row_len_selector == "min":
        n_depth_use = np.min(row_lengths)
    elif row_len_selector == "max":
        n_depth_use = np.max(row_lengths)
    elif row_len_selector == "median":
        n_depth_use = np.median(row_lengths)
        # If the median is half-way between two values, round up
        if n_depth_use not in row_depth_starts:
            n_depth_use = int(np.round(n_depth_use))
        # If the median is still not between values, drop the last value
        # to make the array be odd, guaranteeing the median is an observed
        # value, not an intermediary.
        if n_depth_use not in row_depth_starts:
            n_depth_use = np.median(row_lengths[:-1])
    elif row_len_selector == "mode":
        n_depth_use = utils.mode(row_lengths)
    else:
        raise ValueError(
            "Unsupported row_len_selector value: {}".format(row_len_selector)
        )

    # Use depths corresponding to that declared in the rows which had the
    # number of entries used.
    if row_len_selector == "median":
        d_start = np.median(row_depth_starts[row_lengths == n_depth_use])
        d_stop = np.median(row_depth_ends[row_lengths == n_depth_use])
    else:
        d_start = utils.mode(row_depth_starts[row_lengths == n_depth_use])
        d_stop = utils.mode(row_depth_ends[row_lengths == n_depth_use])
    depths = np.linspace(d_start, d_stop, n_depth_use)

    # Interpolate depths to get a consistent sampling grid
    interp_kwargs = dict(nan_threshold=0.3, assume_sorted=True)
    for i_entry, (nd, d0, d1) in enumerate(
        zip(row_lengths, row_depth_starts, row_depth_ends)
    ):
        if d0 < d1:
            data[i_entry, :n_depth_use] = utils.interp1d_preserve_nan(
                np.linspace(d0, d1, nd),
                data[i_entry, :nd],
                depths,
                **interp_kwargs,
            )
        else:
            data[i_entry, :n_depth_use] = utils.interp1d_preserve_nan(
                np.linspace(d1, d0, nd),
                data[i_entry, :nd][::-1],
                depths,
                **interp_kwargs,
            )

    # Crop the data down to size
    data = data[:, :n_depth_use]

    return timestamps, depths, data


def evl_reader(fname):
    """
    EVL file reader.

    Parameters
    ----------
    fname : str
        Path to .evl file.

    Returns
    -------
    generator
        A generator which yields the timestamp (in seconds), depth (in
        metres), and status (int) for each entry. Note that the timestamp is
        not corrected for timezone (so make sure your timezones are internally
        consistent).
    """
    with open(fname, "r") as hf:
        continuance = True
        for i_row, row in enumerate(csv.reader(hf, delimiter=" ")):
            if i_row == 0:
                continue
            if len(row) < 4:
                if not continuance:
                    raise ValueError("Trying to skip data after parsing began")
                continue
            continuance = False

            timestamp = evdtstr2timestamp(row[0], row[1])

            if len(row[2]) > 0:
                raise ValueError("row[2] was non-empty: {}".format(row[2]))

            yield timestamp, float(row[3]), int(row[4])


def evl_loader(fname, special_to_nan=True, return_status=False):
    """
    EVL file loader.

    Parameters
    ----------
    fname : str
        Path to .evl file.
    special_to_nan : bool, optional
        Whether to replace the special value, ``-10000.99``, which indicates no
        depth value, with NaN.
        https://support.echoview.com/WebHelp/Reference/File_formats/Export_file_formats/Special_Export_Values.htm

    Returns
    -------
    numpy.ndarray of floats
        Timestamps, in seconds.
    numpy.ndarary of floats
        Depth, in metres.
    numpy.ndarary of ints, optional
        Status codes.
    """
    timestamps = []
    values = []
    statuses = []
    for timestamp, value, status in evl_reader(fname):
        timestamps.append(timestamp)
        values.append(value)
        statuses.append(status)
    timestamps = np.array(timestamps)
    values = np.array(values)
    statuses = np.array(statuses)
    if special_to_nan:
        # Replace the special value -10000.99 with NaN
        # https://support.echoview.com/WebHelp/Reference/File_formats/Export_file_formats/Special_Export_Values.htm
        values[np.isclose(values, -10000.99)] = np.nan
    if return_status:
        return timestamps, values, statuses
    return timestamps, values


def timestamp2evdtstr(timestamp):
    """
    Convert a timestamp into an Echoview-compatible datetime string.

    The output is in the format "CCYYMMDD HHmmSSssss", where:

    | CC: century
    | YY: year
    | MM: month
    | DD: day
    | HH: hour
    | mm: minute
    | SS: second
    | ssss: 0.1 milliseconds

    Parameters
    ----------
    timestamp : float
        Number of seconds since Unix epoch.

    Returns
    -------
    datetimestring : str
        Datetime string in the Echoview-compatible format
        "CCYYMMDD HHmmSSssss".
    """
    # Datetime must be in the format CCYYMMDD HHmmSSssss
    # where ssss = 0.1 milliseconds.
    # We have to manually determine the number of "0.1 milliseconds"
    # from the microsecond component.
    dt = datetime.datetime.fromtimestamp(timestamp)
    return "{}{:04d}".format(dt.strftime("%Y%m%d %H%M%S"), round(dt.microsecond / 100))


def evdtstr2timestamp(datestr, timestr=None):
    """
    Convert an Echoview-compatible datetime string into a Unix epoch timestamp.

    Parameters
    ----------
    datestr : str
        Datetime string in the Echoview-compatible format
        ``"CCYYMMDD HHmmSSssss"``, or (if timestr is also provided) just
        the date part, ``"CCYYMMDD"``.
    timestr : str, optional
        Time string in the Echoview-compatible format "HHmmSSssss".

    Returns
    -------
    timestamp : float
        Number of seconds since Unix epoch.
    """
    if timestr:
        datestr = datestr + " " + timestr
    return datetime.datetime.strptime(datestr, "%Y%m%d %H%M%S%f").timestamp()


def evr_reader(fname, parse_echofilter_regions=True):
    """
    Echoview region file (EVR) reader.

    Parameters
    ----------
    fname : str
        Path to .evr file.
    parse_echofilter_regions : bool, default=True
        Whether to separate out echofilter generated regions
        (passive, removed vbands, and removed patches)
        from other regions.

    Returns
    -------
    regions_passive : list of tuples, optional
        Start and end timestamps for passive regions.
    regions_removed : list of tuples, optional
        Start and end timestamps for removed vertical bands.
    regions_patch : list of lists, optional
        Start and end timestamps for bad data patches.
    regions_other : list of dicts
        Dictionary mapping creation type to points defining each region.
    """
    regions_passive = []
    regions_removed = []
    regions_patch = []
    regions_other = []

    # Line 1: EVRG file version header
    with open(fname, "r") as hf:
        line = hf.readline()
        if "EVRG" not in line:
            raise EnvironmentError("This is not an EVR/EVRG file")

        # Line 2: Number of regions
        line = hf.readline().strip("\n\r")
        n_regions = int(line)

        for _ in range(n_regions):
            # Line 3: Intentionally left blank
            line = hf.readline().strip("\n\r")
            if len(line):
                print(
                    "Badly formatted EVRG file. Separating line is not blank: {}".format(
                        line
                    )
                )

            # Individual region
            region_is_passive = False
            region_is_removed = False
            region_is_patch = False

            # Region header
            line = hf.readline().strip("\n\r")
            line = line.replace("  ", " ")
            lparts = line.split(" ")
            if lparts[0] != "13":
                raise EnvironmentError(
                    "Only Region structure version 13 is supported ({} given).".format(
                        lparts[0]
                    )
                )
            n_points = int(lparts[1])
            _region_id = int(lparts[2])  # noqa: F841
            # Selected indicator: lparts[3] == 0
            region_ctype = int(lparts[4])
            # Dummy: lparts[5] == -1
            _has_bbox = bool(int(lparts[6]))  # noqa: F841
            _left = evdtstr2timestamp(lparts[7], lparts[8])  # noqa: F841
            _top = float(lparts[9])  # noqa: F841
            _right = evdtstr2timestamp(lparts[10], lparts[11])  # noqa: F841
            _bottom = float(lparts[12])  # noqa: F841

            # Notes
            line = hf.readline().strip("\n\r")
            n_notes = int(line)
            notes = "\n".join(hf.readline() for _ in range(n_notes))
            if notes.startswith("Passive data"):
                region_is_passive = True
            if notes.startswith("Removed data block"):
                region_is_removed = True
            if notes.startswith("Removed patch"):
                region_is_patch = True

            # Detection settings
            line = hf.readline().strip("\n\r")
            n_detset = int(line)
            for _ in range(n_detset):
                hf.readline()

            # Region classification
            _region_classification = hf.readline().strip("\n\r")  # noqa: F841
            # Points
            points = hf.readline().strip().replace("  ", " ").split(" ")
            _region_status = int(points.pop())  # noqa: F841
            if len(points) % 3 != 0:
                print("Points not composed correctly")
            if len(points) // 3 != n_points:
                print(
                    "Different number of points to expected: {} vs {}".format(
                        len(points), n_points
                    )
                )
            points = [
                (
                    evdtstr2timestamp(points[i * 3], points[i * 3 + 1]),
                    float(points[i * 3 + 2]),
                )
                for i in range(n_points)
            ]
            _region_name = hf.readline().strip("\n\r")  # noqa: F841

            # Add region to list
            if not parse_echofilter_regions:
                regions_other.append({region_ctype: points})
                continue

            if region_is_passive:
                if region_ctype != 4:
                    print(
                        "Warning: Region creation type is {} not 4".format(region_ctype)
                    )
                regions_passive.append((points[0][0], points[-1][0]))
            elif region_is_removed:
                if region_ctype != 4:
                    print(
                        "Warning: Region creation type is {} not 4".format(region_ctype)
                    )
                regions_removed.append((points[0][0], points[-1][0]))
            elif region_is_patch:
                if region_ctype != 2:
                    print(
                        "Warning: Region creation type is {} not 2".format(region_ctype)
                    )
                regions_patch.append(points)
            else:
                regions_other.append({region_ctype: points})

    if not parse_echofilter_regions:
        return regions_other

    return regions_passive, regions_removed, regions_patch, regions_other


def regions2mask(
    timestamps,
    depths,
    regions_passive=None,
    regions_removed=None,
    regions_patch=None,
    regions_other=None,
):
    """
    Convert regions to mask.

    Takes the output from :func:evr_reader` and returns a set of masks.

    Parameters
    ----------
    timestamps : array_like
        Timestamps for each node in the line.
    depths : array_like
        Depths (in meters) for each node in the line.
    regions_passive : list of tuples, optional
        Start and end timestamps for passive regions.
    regions_removed : list of tuples, optional
        Start and end timestamps for removed vertical bands.
    regions_patch : list of lists, optional
        Start and end timestamps for bad data patches.
    regions_other : list of dicts
        Dictionary mapping creation type to points defining each region.

    Returns
    -------
    transect : dict
        A dictionary with keys:

            - "is_passive" : numpy.ndarray
                Logical array showing whether a timepoint is of passive data.
                Shaped ``(num_timestamps, )``. All passive recording data should
                be excluded by the mask.
            - "is_removed" : numpy.ndarray
                Logical array showing whether a timepoint is entirely removed
                by the mask. Shaped ``(num_timestamps, )``.
            - "mask_patches" : numpy.ndarray
                Logical array indicating which datapoints are inside a patch
                from regions_patch (``True``) and should be excluded by the
                mask.
                Shaped ``(num_timestamps, num_depths)``.
            - "mask" : numpy.ndarray
                Logical array indicating which datapoints should be kept
                (``True``) and which are marked as removed
                (``False``) by one of the other three outputs.
                Shaped ``(num_timestamps, num_depths)``.
    """
    if regions_other is not None:
        raise NotImplementedError("Other regions are not yet supported.")

    is_passive = np.zeros(timestamps.shape, dtype=bool)
    is_removed = np.zeros(timestamps.shape, dtype=bool)

    for ts_start, ts_end in regions_passive:
        is_passive[(ts_start <= timestamps) & (timestamps <= ts_end)] = 1

    for ts_start, ts_end in regions_removed:
        is_removed[(ts_start <= timestamps) & (timestamps <= ts_end)] = 1

    # Create an empty image to store the masked array
    mask_patches = np.zeros((len(timestamps), len(depths)), dtype=bool)

    dt = np.abs(timestamps[-1] - timestamps[-2])
    dd = np.abs(depths[-1] - depths[-2])

    for patch in regions_patch:
        patch = np.asarray(patch)
        # Find closest indices to the contour coordinates given
        x = np.searchsorted(timestamps + dt / 2, patch[:, 0], side="left")
        y = np.searchsorted(depths + dd / 2, patch[:, 1], side="left")

        # Make a mask for this patch
        mask_i = np.zeros_like(mask_patches, dtype="bool")

        # Create a contour image by using the contour coordinates rounded to their nearest integer value
        mask_i[x, y] = 1

        # Fill in the hole created by the contour boundary
        mask_i = scipy.ndimage.binary_fill_holes(mask_i)

        # Add this patch to the overall mask for all patches
        mask_patches |= mask_i

    # mask_patches shows where bad patches are; mask shows where good data is
    mask = ~mask_patches
    mask[is_passive | is_removed] = 0

    transect = {
        "is_passive": is_passive,
        "is_removed": is_removed,
        "mask": mask,
        "mask_patches": mask_patches,
    }
    return transect


def evl_writer(fname, timestamps, depths, status=1, line_ending="\r\n", pad=False):
    r"""
    EVL file writer.

    Parameters
    ----------
    fname : str
        Destination of output file.
    timestamps : array_like
        Timestamps for each node in the line.
    depths : array_like
        Depths (in meters) for each node in the line.
    status : 0, 1, 2, or 3; optional
        Status for the line.

        - ``0`` : none
        - ``1`` : unverified
        - ``2`` : bad
        - ``3`` : good

        Default is ``1`` (unverified). For more details on line status, see
        https://support.echoview.com/WebHelp/Using_Echoview/Echogram/Lines/About_Line_Status.htm
    pad : bool, optional
        Whether to pad the line with an extra datapoint half a pixel before the
        first and after the last given timestamp. Default is ``False``.
    line_ending : str, optional
        Line ending. Default is ``"\r\n"`` the standard line ending on Windows/DOS,
        as per the specification for the file format.
        https://support.echoview.com/WebHelp/Using_Echoview/Exporting/Exporting_data/Exporting_line_data.htm
        Set to ``"\n"`` to get Unix-style line endings instead.

    Notes
    -----
    For more details on the format specification, see
    https://support.echoview.com/WebHelp/Using_Echoview/Exporting/Exporting_data/Exporting_line_data.htm#Line_definition_file_format
    """
    if len(timestamps) != len(depths):
        raise ValueError(
            "Number of timestamps ({}) and depths ({}) are not equal".format(
                len(timestamps), len(depths)
            )
        )
    if pad and len(timestamps) > 1:
        timestamps = timestamps[:]
        timestamps = np.r_[
            timestamps[0] - (timestamps[1] - timestamps[0]) / 2,
            timestamps,
            timestamps[-1] + (timestamps[-2] - timestamps[-1]) / 2,
        ]
        depths = np.r_[depths[0], depths, depths[-1]]
    # The file object will automatically replace \n with our chosen line ending
    with open(fname, "w+", encoding="utf-8-sig", newline=line_ending) as hf:
        # Write header
        hf.write("EVBD 3 10.0.270.37090\n")
        n_row = len(depths)
        hf.write(str(n_row) + "\n")
        # Write each row
        for timestamp, depth in zip(timestamps, depths):
            # Datetime must be in the format CCYYMMDD HHmmSSssss
            # where ssss = 0.1 milliseconds.
            # We have to manually determine the number of "0.1 milliseconds"
            # from the microsecond component.
            hf.write("{}  {} {} \n".format(timestamp2evdtstr(timestamp), depth, status))


def evr_writer(
    fname,
    rectangles=None,
    contours=None,
    common_notes="",
    default_region_type=0,
    line_ending="\r\n",
):
    r"""
    EVR file writer.

    Writes regions to an Echoview region file.

    Parameters
    ----------
    fname : str
        Destination of output file.
    rectangles : list of dictionaries, optional
        Rectangle region definitions. Default is an empty list. Each rectangle
        region must implement fields ``"depths"`` and ``"timestamps"``, which
        indicate the extent of the rectangle. Optionally, ``"creation_type"``,
        ``"region_name"``, ``"region_type"``, and ``"notes"`` may be set.
        If these are not given, the default creation_type is 4 and region_type
        is set by ``default_region_type``.
    contours : list of dictionaries
        Contour region definitions. Default is an empty list. Each contour
        region must implement a ``"points"`` field containing a :class:`numpy.ndarray`
        shaped `(n, 2)` defining the co-ordinates of nodes along the (open)
        contour in units of timestamp and depth. Optionally, ``"creation_type"``,
        ``"region_name"``, ``"region_type"``, and ``"notes"`` may be set.
        If these are not given, the default creation_type is 2 and region_type
        is set by ``default_region_type``.
    common_notes : str, optional
        Notes to include for every region. Default is ``""``, an empty string.
    default_region_type : int, optional
        The region type to use for rectangles and contours which do not define
        a ``"region_type"`` field. Possible region types are

        - ``0`` : bad (no data)
        - ``1`` : analysis
        - ``2`` : marker
        - ``3`` : fishtracks
        - ``4`` : bad (empty water)

        Default is ``0``.
    line_ending : str, optional
        Line ending. Default is ``"\r\n"`` the standard line ending on Windows/DOS,
        as per the specification for the file format.
        https://support.echoview.com/WebHelp/Using_Echoview/Exporting/Exporting_data/Exporting_line_data.htm
        Set to ``"\n"`` to get Unix-style line endings instead.

    Notes
    -----
    For more details on the format specification, see:
    https://support.echoview.com/WebHelp/Reference/File_formats/Export_file_formats/2D_Region_definition_file_format.htm
    """
    if rectangles is None:
        rectangles = []
    if contours is None:
        contours = []
    # Remove leading/trailing new lines, since we will join with our own line ending
    common_notes = common_notes.strip("\r\n")
    # Standardize line endings to be \n, regardless of input
    common_notes = common_notes.replace("\r\n", "\n").replace("\r", "\n")
    if len(common_notes) == 0:
        n_lines_common_notes = 0
    else:
        n_lines_common_notes = 1 + common_notes.count(line_ending)
    n_regions = len(rectangles) + len(contours)
    i_region = 0
    # The file object will automatically replace \n with our chosen line ending
    with open(fname, "w+", encoding="utf-8-sig", newline=line_ending) as hf:
        # Write header
        hf.write("EVRG 7 10.0.283.37689\n")
        hf.write(str(n_regions) + "\n")

        # Write each rectangle
        for region in rectangles:
            # Regions are indexed from 1, so increment the counter first
            i_region += 1
            hf.write("\n")  # Blank line separates regions
            # Determine extent of rectangle
            left = timestamp2evdtstr(np.min(region["timestamps"]))
            right = timestamp2evdtstr(np.max(region["timestamps"]))
            top = np.min(region["depths"])
            bottom = np.max(region["depths"])
            # Region header
            hf.write(
                "13 4 {i} 0 {type} -1 1 {left}  {top} {right}  {bottom}".format(
                    i=i_region,
                    type=region.get("creation_type", 4),
                    left=left,
                    right=right,
                    top=top,
                    bottom=bottom,
                )
                + "\n"
            )
            # Notes
            notes = region.get("notes", "")
            if len(notes) == 0:
                notes = common_notes
                n_lines_notes = n_lines_common_notes
            else:
                notes = notes.strip("\n")
                if len(common_notes) > 0:
                    notes += "\n" + common_notes
                n_lines_notes = 1 + notes.count("\n")
            hf.write(str(n_lines_notes) + "\n")  # Number of lines of notes
            if len(notes) > 0:
                hf.write(notes + "\n")
            # Detection settings
            hf.write("0\n")  # Number of lines of detection settings
            # Region classification string
            hf.write("Unclassified regions\n")
            # The points defining the region itself
            hf.write(
                "{left} {top} {left} {bottom} {right} {bottom} {right} {top} ".format(
                    left=left,
                    right=right,
                    top=top,
                    bottom=bottom,
                )  # Terminates with a space, not a new line
            )
            # Region type
            hf.write(str(region.get("region_type", default_region_type)) + "\n")
            # Region name
            hf.write(
                str(region.get("region_name", "Region {}".format(i_region))) + "\n"
            )

        # Write each contour
        for region in contours:
            # Regions are indexed from 1, so increment the counter first
            i_region += 1
            hf.write("\n")  # Blank line separates regions
            # Header line
            hf.write(
                "13 {n} {i} 0 {type} -1 1 {left}  {top} {right}  {bottom}".format(
                    n=region["points"].shape[0],
                    i=i_region,
                    type=region.get("creation_type", 2),
                    left=timestamp2evdtstr(np.min(region["points"][:, 0])),
                    right=timestamp2evdtstr(np.max(region["points"][:, 0])),
                    top=np.min(region["points"][:, 1]),
                    bottom=np.max(region["points"][:, 1]),
                )
                + "\n"
            )
            # Notes
            notes = region.get("notes", "")
            if len(notes) == 0:
                notes = common_notes
                n_lines_notes = n_lines_common_notes
            else:
                notes = notes.strip("\n")
                if len(common_notes) > 0:
                    notes += "\n" + common_notes
                n_lines_notes = 1 + notes.count("\n")
            hf.write(str(n_lines_notes) + "\n")  # Number of lines of notes
            if len(notes) > 0:
                hf.write(notes + "\n")
            # Detection settings
            hf.write("0\n")  # Number of lines of detection settings
            # Region classification string
            hf.write("Unclassified regions\n")
            # The region itself
            for point in region["points"]:
                hf.write("{} {} ".format(timestamp2evdtstr(point[0]), point[1]))
            # Region type
            hf.write(str(region.get("region_type", default_region_type)) + "\n")
            # Region name
            hf.write(
                str(region.get("region_name", "Region {}".format(i_region))) + "\n"
            )


def write_transect_regions(
    fname,
    transect,
    depth_range=None,
    passive_key="is_passive",
    removed_key="is_removed",
    patches_key="mask_patches",
    collate_passive_length=0,
    collate_removed_length=0,
    minimum_passive_length=0,
    minimum_removed_length=0,
    minimum_patch_area=0,
    name_suffix="",
    common_notes="",
    line_ending="\r\n",
    verbose=0,
    verbose_indent=0,
):
    r"""
    Convert a transect dictionary to a set of regions and write as an EVR file.

    Parameters
    ----------
    fname : str
        Destination of output file.
    transect : dict
        Transect dictionary.
    depth_range : array_like or None, optional
        The minimum and maximum depth extents (in any order) of the passive and
        removed block regions. If this is ``None`` (default), the minimum and
        maximum of ``transect["depths"]`` is used.
    passive_key : str, optional
        Field name to use for passive data identification. Default is
        ``"is_passive"``.
    removed_key : str, optional
        Field name to use for removed blocks. Default is ``"is_removed"``.
    patches_key : str, optional
        Field name to use for the mask of patch regions. Default is
        ``"mask_patches"``.
    collate_passive_length : int, optional
        Maximum distance (in indices) over which passive regions should be
        merged together, closing small gaps between them. Default is ``0``.
    collate_removed_length : int, optional
        Maximum distance (in indices) over which removed blocks should be
        merged together, closing small gaps between them. Default is ``0``.
    minimum_passive_length : int, optional
        Minimum length (in indices) a passive region must have to be included
        in the output. Set to -1 to omit all passive regions from the output.
        Default is ``0``.
    minimum_removed_length : int, optional
        Minimum length (in indices) a removed block must have to be included in
        the output. Set to -1 to omit all removed regions from the output.
        Default is ``0``.
    minimum_patch_area : float, optional
        Minimum amount of area (in input pixel space) that a patch must occupy
        in order to be included in the output. Set to ``0`` to include all
        patches, no matter their area. Set to ``-1`` to omit all patches.
        Default is ``0``.
    name_suffix : str, optional
        Suffix to append to variable names. Default is ``""``, an empty string.
    common_notes : str, optional
        Notes to include for every region. Default is ``""``, an empty string.
    line_ending : str, optional
        Line ending. Default is ``"\r\n"`` the standard line ending on Windows/DOS,
        as per the specification for the file format,
        https://support.echoview.com/WebHelp/Using_Echoview/Exporting/Exporting_data/Exporting_line_data.htm
        Set to ``"\n"`` to get Unix-style line endings instead.
    verbose : int, optional
        Verbosity level. Default is ``0``.
    verbose_indent : int, optional
        Level of indentation (number of preceding spaces) before verbosity
        messages. Default is ``0``.
    """
    if depth_range is None:
        depth_range = transect["depths"]
    depth_range = [np.min(depth_range), np.max(depth_range)]

    rectangles = []
    # Regions around each period of passive data
    key = passive_key
    if key not in transect:
        key = "p_" + key
    if key not in transect:
        raise ValueError("Key {} and {} not found in transect.".format(key[2:], key))
    is_passive = transect[key] > 0.5
    is_passive = ~utils.squash_gaps(~is_passive, collate_passive_length)
    passive_starts, passive_ends = utils.get_indicator_onoffsets(is_passive)
    i_passive = 1
    n_passive_skipped = 0
    for start_index, end_index in zip(passive_starts, passive_ends):
        start_index -= 0.5
        end_index += 0.5
        if minimum_passive_length == -1:
            # No passive regions
            break
        if end_index - start_index <= minimum_passive_length:
            n_passive_skipped += 1
            continue
        region = {}
        region["region_name"] = "Passive{} {}".format(name_suffix, i_passive)
        region["creation_type"] = 4
        region["region_type"] = 0
        region["depths"] = depth_range
        region["timestamps"] = scipy.interpolate.interp1d(
            np.arange(len(transect["timestamps"])),
            transect["timestamps"],
            fill_value="extrapolate",
        )([start_index, end_index])
        region["notes"] = textwrap.dedent(
            """
            Passive data
            Length in pixels: {}
            Duration in seconds: {}
            """.format(
                end_index - start_index,
                region["timestamps"][1] - region["timestamps"][0],
            )
        )
        rectangles.append(region)
        i_passive += 1
    # Regions around each period of removed data
    key = removed_key
    if key not in transect:
        key = "p_" + key
    if key not in transect:
        raise ValueError("Key {} and {} not found in transect.".format(key[2:], key))
    is_removed = transect[key] > 0.5
    is_removed = ~utils.squash_gaps(~is_removed, collate_removed_length)
    removed_starts, removed_ends = utils.get_indicator_onoffsets(is_removed)
    i_removed = 1
    n_removed_skipped = 0
    for start_index, end_index in zip(removed_starts, removed_ends):
        start_index -= 0.5
        end_index += 0.5
        if minimum_removed_length == -1:
            # No passive regions
            break
        if end_index - start_index <= minimum_removed_length:
            n_removed_skipped += 1
            continue
        region = {}
        region["region_name"] = "Removed block{} {}".format(name_suffix, i_removed)
        region["creation_type"] = 4
        region["region_type"] = 0
        region["depths"] = depth_range
        region["timestamps"] = scipy.interpolate.interp1d(
            np.arange(len(transect["timestamps"])),
            transect["timestamps"],
            fill_value="extrapolate",
        )([start_index, end_index])
        region["notes"] = textwrap.dedent(
            """
            Removed data block
            Length in pixels: {}
            Duration in seconds: {}
            """.format(
                end_index - start_index,
                region["timestamps"][1] - region["timestamps"][0],
            )
        )
        rectangles.append(region)
        i_removed += 1
    # Contours around each removed patch
    if patches_key not in transect:
        raise ValueError("Key {} not found in transect.".format(patches_key))
    patches = transect[patches_key]
    patches = scipy.ndimage.binary_fill_holes(patches > 0.5)
    contours_coords = skimage.measure.find_contours(patches, 0.5)
    contour_dicts = []
    i_contour = 1
    n_contour_skipped = 0
    for contour in contours_coords:
        if minimum_patch_area == -1:
            # No patches
            break
        area = utils.integrate_area_of_contour(
            contour[:, 0], contour[:, 1], closed=False
        )
        if area < minimum_patch_area:
            n_contour_skipped += 1
            continue
        region = {}
        region["region_name"] = "Removed patch{} {}".format(name_suffix, i_contour)
        region["creation_type"] = 2
        region["region_type"] = 0
        x = scipy.interpolate.interp1d(
            np.arange(len(transect["timestamps"])),
            transect["timestamps"],
            fill_value="extrapolate",
        )(contour[:, 0])
        y = scipy.interpolate.interp1d(
            np.arange(len(transect["depths"])),
            transect["depths"],
            fill_value="extrapolate",
        )(contour[:, 1])
        region["points"] = np.stack([x, y], axis=-1)
        region["notes"] = textwrap.dedent(
            """
            Removed patch
            Area in pixels: {}
            Area in meter-seconds: {}
            """.format(
                area, utils.integrate_area_of_contour(x, y, closed=False)
            )
        )
        contour_dicts.append(region)
        i_contour += 1
    if verbose >= 1:
        print(
            " " * verbose_indent + "Outputting {} region{}:"
            " {} passive, {} removed blocks, {} removed patches".format(
                len(rectangles) + len(contour_dicts),
                "" if len(rectangles) + len(contour_dicts) == 1 else "s",
                i_passive - 1,
                i_removed - 1,
                i_contour - 1,
            )
        )
        n_skipped = n_passive_skipped + n_removed_skipped + n_contour_skipped
        if n_skipped > 0:
            print(
                " " * verbose_indent
                + style.skip_fmt(
                    "There {} {} skipped (too small) region{}:"
                    " {} passive, {} removed blocks, {} removed patches".format(
                        "was" if n_skipped == 1 else "were",
                        n_skipped,
                        "" if n_skipped == 1 else "s",
                        n_passive_skipped,
                        n_removed_skipped,
                        n_contour_skipped,
                    )
                )
            )

    # Write the output
    return evr_writer(
        fname,
        rectangles=rectangles,
        contours=contour_dicts,
        common_notes=common_notes,
        line_ending=line_ending,
    )


def load_transect_data(transect_pth, dataset="mobile", root_data_dir=ROOT_DATA_DIR):
    """
    Load all data for one transect.

    Parameters
    ----------
    transect_pth : str
        Relative path to transect, excluding ``"_Sv_raw.csv"``.
    dataset : str, optional
        Name of dataset. Default is ``"mobile"``.
    root_data_dir : str
        Path to root directory where data is located.

    Returns
    -------
    timestamps : numpy.ndarray
        Timestamps (in seconds since Unix epoch), with each entry
        corresponding to each row in the ``signals`` data.
    depths : numpy.ndarray
        Depths from the surface (in metres), with each entry corresponding
        to each column in the ``signals`` data.
    signals : numpy.ndarray
        Echogram Sv data, shaped (num_timestamps, num_depths).
    turbulence : numpy.ndarray
        Depth of turbulence line, shaped (num_timestamps, ).
    bottom : numpy.ndarray
        Depth of bottom line, shaped (num_timestamps, ).
    """
    dirname = os.path.join(root_data_dir, dataset)
    raw_fname = os.path.join(dirname, transect_pth + "_Sv_raw.csv")
    bottom_fname = os.path.join(dirname, transect_pth + "_bottom.evl")
    turbulence_fname = os.path.join(dirname, transect_pth + "_turbulence.evl")

    timestamps, depths, signals = transect_loader(raw_fname)
    t_bottom, d_bottom = evl_loader(bottom_fname)
    t_turbulence, d_turbulence = evl_loader(turbulence_fname)

    return (
        timestamps,
        depths,
        signals,
        np.interp(timestamps, t_turbulence, d_turbulence),
        np.interp(timestamps, t_bottom, d_bottom),
    )


def get_partition_data(
    partition,
    dataset="mobile",
    partitioning_version="firstpass",
    root_data_dir=ROOT_DATA_DIR,
):
    """
    Load partition metadata.

    Parameters
    ----------
    transect_pth : str
        Relative path to transect, excluding ``"_Sv_raw.csv"``.
    dataset : str, optional
        Name of dataset. Default is ``"mobile"``.
    partitioning_version : str, optional
        Name of partitioning method.
    root_data_dir : str
        Path to root directory where data is located.

    Returns
    -------
    pandas.DataFrame
        Metadata for all transects in the partition. Each row is a single
        sample.
    """
    dirname = os.path.join(root_data_dir, dataset, "sets", partitioning_version)
    fname_partition = os.path.join(dirname, partition + ".txt")
    fname_header = os.path.join(dirname, "header.txt")

    with open(fname_header, "r") as hf:
        for row in csv.reader(hf):
            header = [entry.strip() for entry in row]
            break

    df = pd.read_csv(fname_partition, header=None, names=header)
    return df


def remove_trailing_slash(s):
    """
    Remove trailing forward slashes from a string.

    Parameters
    ----------
    s : str
        String representing a path, possibly with trailing slashes.

    Returns
    -------
    str
        Same as ``s``, but without trailing forward slashes.
    """
    while s[-1] == "/" or s[-1] == os.path.sep:
        s = s[:-1]
    return s


def list_from_file(fname):
    """
    Get a list from a file.

    Parameters
    ----------
    fname : str
        Path to file.

    Returns
    -------
    list
        Contents of the file, one line per entry in the list. Trailing
        whitespace is removed from each end of each line.
    """
    with open(fname, "r") as hf:
        contents = hf.readlines()
    contents = [x.strip() for x in contents]
    return contents


def get_partition_list(
    partition,
    dataset="mobile",
    full_path=False,
    partitioning_version="firstpass",
    root_data_dir=ROOT_DATA_DIR,
    sharded=False,
):
    """
    Get a list of transects in a single partition.

    Parameters
    ----------
    transect_pth : str
        Relative path to transect, excluding ``"_Sv_raw.csv"``.
    dataset : str, optional
        Name of dataset. Default is ``"mobile"``.
    full_path : bool, optional
        Whether to return the full path to the sample. If ``False``, only the
        relative path (from the dataset directory) is returned.
        Default is ``False``.
    partitioning_version : str, optional
        Name of partitioning method.
    root_data_dir : str, optional
        Path to root directory where data is located.
    sharded : bool, optional
        Whether to return path to sharded version of data. Default is ``False``.

    Returns
    -------
    list
        Path for each sample in the partition.
    """
    if dataset == "mobile":
        df = get_partition_data(
            partition,
            dataset=dataset,
            partitioning_version=partitioning_version,
            root_data_dir=root_data_dir,
        )
        fnames = df["Filename"]
        fnames = [os.path.join(f.split("_")[0], f.strip()) for f in fnames]
    else:
        partition_file = os.path.join(
            root_data_dir,
            dataset,
            "sets",
            partitioning_version,
            partition + ".txt",
        )
        fnames = list_from_file(partition_file)

    fnames = [f.replace("_Sv_raw.csv", "") for f in fnames]
    if full_path and sharded:
        root_data_dir = remove_trailing_slash(root_data_dir)
        fnames = [os.path.join(root_data_dir + "_sharded", dataset, f) for f in fnames]
    elif full_path:
        fnames = [os.path.join(root_data_dir, dataset, f) for f in fnames]
    return fnames



"""
Utility functions for dataset.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import random

import numpy as np
import torch

from ..utils import first_nonzero, last_nonzero


def worker_seed_fn(worker_id):
    """
    Seed builtin :mod:`random` and :mod:`numpy` with :meth:`torch.randint`.

    A worker initialization function for :class:`torch.utils.data.DataLoader` objects
    which seeds builtin :mod:`random` and :mod:`numpy` with :meth:`torch.randint` (which is
    stable if torch is manually seeded in the main program).

    Parameters
    ----------
    worker_id : int
        The ID of the worker.
    """
    np.random.seed((torch.randint(0, 4294967296, (1,)).item() + worker_id) % 4294967296)
    random.seed(torch.randint(0, 4294967296, (1,)).item() + worker_id)


def worker_staticseed_fn(worker_id):
    """
    Seed builtin :mod:`random`, :mod:`numpy`, and :mod:`torch` with ``worker_id``.

    A worker initialization function for :class:`torch.utils.data.DataLoader`
    objects which produces the same seed for builtin :mod:`random`, :mod:`numpy`, and
    :mod:`torch` every time, so it is the same for every epoch.

    Parameters
    ----------
    worker_id : int
        The ID of the worker.
    """
    random.seed(worker_id)
    np.random.seed(worker_id)
    torch.manual_seed(worker_id)



"""
Pytorch activation functions.

Swish and Mish implementations taken from https://github.com/fastai/fastai2
under the Apache License Version 2.0.
"""

import functools

import torch
import torch.nn.functional as F
from torch import nn

__all__ = [
    "str2actfnfactory",
    "InplaceReLU",
    "swish",
    "Swish",
    "HardSwish",
    "mish",
    "Mish",
    "HardMish",
]


def str2actfnfactory(actfn_name):
    """
    Map an activation function name to a factory which generates that actfun.

    Parameters
    ----------
    actfn_name : str
        Name of the activation function.

    Returns
    -------
    callable
        A generator which yields a subclass of :class:`torch.nn.Module`.
    """
    if hasattr(nn, actfn_name):
        return getattr(nn, actfn_name)

    actfn_name_og = actfn_name
    actfn_name = actfn_name.lower().replace("-", "").replace("_", "")
    if actfn_name == "inplacerelu" or actfn_name == "reluinplace":
        return InplaceReLU
    elif actfn_name == "swish":
        return Swish
    elif actfn_name == "hardswish":
        return HardSwish
    elif actfn_name == "mish":
        return Mish
    elif actfn_name == "hardmish":
        return HardMish
    else:
        raise ValueError("Unrecognised activation function: {}".format(actfn_name_og))


InplaceReLU = functools.partial(nn.ReLU, inplace=True)


# Swish
@torch.jit.script
def _swish_jit_fwd(x):
    return x.mul(torch.sigmoid(x))


@torch.jit.script
def _swish_jit_bwd(x, grad_output):
    x_sigmoid = torch.sigmoid(x)
    return grad_output * (x_sigmoid * (1 + x * (1 - x_sigmoid)))


class _SwishJitAutoFn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return _swish_jit_fwd(x)

    @staticmethod
    def backward(ctx, grad_output):
        x = ctx.saved_variables[0]
        return _swish_jit_bwd(x, grad_output)


def swish(x, inplace=False):
    return _SwishJitAutoFn.apply(x)


class Swish(nn.Module):
    def forward(self, x):
        return _SwishJitAutoFn.apply(x)


class HardSwish(nn.Module):
    """
    A second-order approximation to the swish activation function.

    See https://arxiv.org/abs/1905.02244
    """

    def __init__(self, inplace=True):
        super().__init__()
        self.inplace = inplace
        self.relu6 = torch.nn.ReLU6(inplace=inplace)

    def forward(self, x):
        return x * self.relu6(x + 3) / 6

    def extra_repr(self):
        inplace_str = "inplace=True" if self.inplace else "inplace=False"
        return inplace_str


# Mish
@torch.jit.script
def _mish_jit_fwd(x):
    return x.mul(torch.tanh(F.softplus(x)))


@torch.jit.script
def _mish_jit_bwd(x, grad_output):
    x_sigmoid = torch.sigmoid(x)
    x_tanh_sp = F.softplus(x).tanh()
    return grad_output.mul(x_tanh_sp + x * x_sigmoid * (1 - x_tanh_sp * x_tanh_sp))


class MishJitAutoFn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return _mish_jit_fwd(x)

    @staticmethod
    def backward(ctx, grad_output):
        x = ctx.saved_variables[0]
        return _mish_jit_bwd(x, grad_output)


def mish(x):
    """
    Apply the mish function elementwise.

    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))

    See https://arxiv.org/abs/1908.08681
    """
    return MishJitAutoFn.apply(x)


class Mish(nn.Module):
    """
    Apply the mish function elementwise.

    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))

    See https://arxiv.org/abs/1908.08681
    """

    def forward(self, x):
        return MishJitAutoFn.apply(x)


class HardMish(nn.Module):
    """
    A second-order approximation to the mish activation function.

    Notes
    -----
    https://forums.fast.ai/t/hard-mish-activation-function/59238
    """

    def __init__(self, inplace=True):
        self.relu5 = nn.Hardtanh(0.0, 5.0, inplace)

    def forward(self, x):
        return x * self.relu5(x + 3) / 5

    def extra_repr(self):
        inplace_str = "inplace=True" if self.inplace else "inplace=False"
        return inplace_str



"""
nn.modules utility functions.
"""

import collections
from itertools import repeat

import torch
from torch import nn

__all__ = ["same_to_padding", "init_cnn"]


def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            x = tuple(x)
            if len(x) == 0:
                raise ValueError("Input {} is an empty iterable".format(x))
            if len(x) > 1 or n == 1:
                return x
            x = x[0]
        return tuple(repeat(x, n))

    return parse


_single = _ntuple(1)
_pair = _ntuple(2)
_triple = _ntuple(3)
_quadruple = _ntuple(4)


def same_to_padding(kernel_size, stride=1, dilation=1, ndim=None):
    """
    Determine the amount of padding to use for a convolutional layer.

    Parameters
    ----------
    kernel_size : int or sequence
        Size of kernel for each dimension.
    stride : int or sequence, optional
        Amount of stride to apply in each dimension of the kernel. If ``stride``
        is an int, the same value is applied for each dimension.
        Default is ``1``.
    dilation : int or sequence, optional
        Amount of dilation to apply in each dimension of the kernel.
        If ``dilation`` is an int, the same value is applied for each dimension.
        Default is ``1``.
    ndim : int or None, optional
        Number of dimensions of kernel to pad. If ``None`` (default), the number
        of dimensions is inferred from the number of dimensions to
        ``kernel_size``.

    Returns
    -------
    padding : tuple
        Amount of padding to apply to each dimension before convolving with
        the kernel in order to preserve the size of input.
    """
    if isinstance(kernel_size, int):
        if kernel_size % 2 == 0:
            raise ValueError("Same padding is not implemented for even kernels sizes.")
    elif any(k % 2 == 0 for k in kernel_size):
        raise ValueError("Same padding is not implemented for even kernels sizes.")

    args = [kernel_size, stride, dilation]

    if all([isinstance(arg, int) for arg in args]):
        return (kernel_size - 1) * dilation // 2

    if ndim is not None:
        pass
    elif isinstance(kernel_size, collections.sequence):
        ndim = len(kernel_size)
    elif isinstance(stride, collections.sequence):
        ndim = len(stride)
    elif isinstance(dilation, collections.sequence):
        ndim = len(dilation)
    else:
        raise ValueError("Wrong argument type. Must be int or iterable.")

    ntuple = _ntuple(ndim)
    kernel_size = ntuple(kernel_size)
    dilation = ntuple(dilation)

    padding = tuple((k - 1) * d // 2 for k, d in zip(kernel_size, dilation))

    return padding


def rint(x, minval=None):
    """
    Round and cast as an int, optionally with a floor value.

    Parameters
    ----------
    x : float
        Number to round.
    minval : bool, optional
        A floor value for the output. If ``None``, no floor is applied. Default
        is ``None``.

    Returns
    -------
    int
        The number rounded to the nearest int, and cast as an int. If ``minval``
        is set, the max with ``minval`` is taken.
    """
    x = int(round(x))
    if minval is not None:
        x = max(minval, x)
    return x


def init_cnn(m):
    """
    Initialize biases and weights for a CNN layer.

    Uses a Kaiming normal distribution for the weight and 0 for biases.

    Function is applied recursively within the module.

    Parameters
    ----------
    m : torch.nn.Module
        Module
    """
    if getattr(m, "bias", None) is not None:
        nn.init.constant_(m.bias, 0)
    if isinstance(m, (nn._ConvNd, nn.Linear)):
        nn.init.kaiming_normal_(m.weight)
    for layer in m.children():
        init_cnn(layer)



"""
Utility functions for interacting with optimizers.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


def get_current_lr(optimizer):
    """
    Get the learning rate of an optimizer.

    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        An optimizer, with a learning rate common to all parameter groups.

    Returns
    -------
    float
        The learning rate of the first parameter group.
    """
    return optimizer.param_groups[0]["lr"]


def get_current_momentum(optimizer):
    """
    Get the momentum of an optimizer.

    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        An optimizer which implements momentum or betas (where momentum is the
        first beta, c.f. :class:`torch.optim.Adam`) with a momentum common to all
        parameter groups.

    Returns
    -------
    float
        The momentum of the first parameter group.
    """
    if "momentum" not in optimizer.defaults and "betas" not in optimizer.defaults:
        raise ValueError(
            "optimizer {} does not support momentum".format(optimizer.__class__)
        )

    group = optimizer.param_groups[0]
    if "momentum" in group:
        return group["momentum"]
    else:
        return group["betas"][0]



"""
Plotting utilities.
"""

import copy

import matplotlib.pyplot as plt
import numpy as np

from . import utils

TURBULENCE_COLOR = "c"
BOTTOM_COLOR = "#00ee00"

# ColorBrewer Paired values
TURBULENCE_COLOR = "#a6cee3"
TURBULENCE_COLOR_DARK = "#1f78b4"
BOTTOM_COLOR = "#b2df8a"
BOTTOM_COLOR_DARK = "#33a02c"
SURFACE_COLOR = "#4ba82a"
SURFACE_COLOR_DARK = "#255e11"

PASSIVE_COLOR = [0.4, 0.4, 0.4]
REMOVED_COLOR = "b"


def ensure_axes_inverted(axes=None, dir="y"):
    """
    Invert axis direction, if not already inverted.

    Parameters
    ----------
    axes : matplotlib.axes or None
        The axes to invert. If ``None``, the current axes are used (default).
    dir : {"x", "y", "xy"}
        The axis to invert. Default is ``"y"``.
    """
    if axes is None:
        axes = plt.gca()
    if "y" in dir:
        lim = axes.get_ylim()
        if lim[1] > lim[0]:
            axes.invert_yaxis()
    if "x" in dir:
        lim = axes.get_xlim()
        if lim[1] > lim[0]:
            axes.invert_xaxis()


def plot_indicator_hatch(
    indicator, xx=None, ymin=None, ymax=None, hatch="//", color="k"
):
    """
    Plot a hatch across indicated segments along the x-axis of a plot.

    Parameters
    ----------
    indicator : numpy.ndarray vector
        Whether to include or exclude each column along the x-axis. Included
        columns are indicated with non-zero values.
    xx : numpy.ndarray vector, optional
        Values taken by indicator along the x-axis. If ``None`` (default), the
        indices of ``indicator`` are used: ``arange(len(indicator))``.
    ymin : float, optional
        The lower y-value of the extent of the hatching. If ``None`` (default),
        the minimum y-value of the current axes is used.
    ymax : float, optional
        The upper y-value of the extent of the hatching. If ``None`` (default),
        the maximum y-value of the current axes is used.
    hatch : str, optional
        Hatching pattern to use. Default is ``"//"``.
    color : color, optional
        Color of the hatching pattern. Default is black.
    """
    if xx is None:
        xx = np.arange(len(indicator))
    if ymin is None or ymax is None:
        ylim = plt.gca().get_ylim()
    if ymin is None:
        ymin = ylim[0]
    if ymax is None:
        ymax = ylim[1]

    r_starts, r_ends = utils.get_indicator_onoffsets(indicator)

    if len(r_starts) == 0:
        return

    for r_start, r_end in zip(r_starts, r_ends):
        plt.fill_between(
            xx[[r_start, r_end]],
            [ymin, ymin],
            [ymax, ymax],
            facecolor="none",
            hatch=hatch,
            edgecolor=color,
            linewidth=0.0,
        )


def plot_mask_hatch(*args, hatch="//", color="k", border=False):
    """
    Plot hatching according to a mask shape.

    Parameters
    ----------
    X, Y : array-like, optional
        The coordinates of the values in Z.

        X and Y must both be 2-D with the same shape as ``Z`` (e.g. created via
        numpy.meshgrid), or they must both be 1-D such that ``len(X) == M`` is
        the number of columns in ``Z`` and ``len(Y) == N`` is the number of rows
        in ``Z``.

        If not given, they are assumed to be integer indices, i.e.
        ``X = range(M)``, ``Y = range(N)``.
    Z : array-like(N, M)
        Indicator for which locations should be hatched. If ``Z`` is not a boolean
        array, any location where ``Z > 0`` will be hatched.
    hatch : str, optional
        The hatching pattern to apply. Default is "//".
    color : color, optional
        The color of the hatch. Default is black.
    border : bool, optional
        Whether to include border around hatch. Default is ``False``.
    """
    args = list(args)
    args[-1] = args[-1] > 0
    if len(args) == 3:
        # Transpose Z if necessary, as it expects its shape the opposite way
        # around to what you might expect.
        if (
            args[0].shape[0] == args[2].shape[1]
            and args[1].shape[0] == args[2].shape[0]
        ):
            pass
        elif (
            args[0].shape[0] == args[2].shape[0]
            and args[1].shape[0] == args[2].shape[1]
        ):
            args[2] = args[2].T
    cs = plt.contourf(
        *args,
        levels=[-3.4e38, 0.5, 3.4e38],
        colors=["none", "none"],
        hatches=[None, hatch],
    )
    # Change the color of the hatches in each layer
    for collection in cs.collections:
        collection.set_edgecolor(color)
    # Doing this also colors in the box around each level.
    # We can remove the colored line around the levels by setting the
    # linewidth to 0.
    if not border:
        for collection in cs.collections:
            collection.set_linewidth(0.0)


def plot_transect(
    transect,
    signal_type=None,
    x_scale="index",
    show_regions=True,
    turbulence_color=TURBULENCE_COLOR,
    bottom_color=BOTTOM_COLOR,
    surface_color=SURFACE_COLOR,
    passive_color=PASSIVE_COLOR,
    removed_color=None,
    linewidth=1,
    cmap=None,
):
    r"""
    Plot a transect.

    Parameters
    ----------
    transect : dict
        Transect values.
    signal_type : str, optional
        The signal to plot as a heatmap. Default is ``"Sv"`` if present, or
        "signals" if not. If this is ``"Sv_masked"``, the mask (given by
        ``transect["mask"]``) is used to mask ``transect["Sv"]`` before plotting.
    x_scale : {"index", "timestamp" "time"}, optional
        Scaling for x-axis. If ``"timestamp"``, the number of seconds since the
        Unix epoch is shown; if ``"time"``, the amount of time in seconds since
        the start of the transect is shown. Default is ``"index"``.
    show_regions : bool, optional
        Whether to show segments of data maked as removed or passive with
        hatching. Passive data is shown with ``"/"`` oriented lines, other removed
        timestamps with ``"\"`` oriented lines. Default is ``True``.
    turbulence_color : color, optional
        Color of turbulence line. Default is ``"#a6cee3"``.
    bottom_color : color, optional
        Color of bottom line. Default is ``"#b2df8a"``.
    surface_color : color, optional
        Color of surface line. Default is ``"#d68ade"``.
    passive_color : color, optional
        Color of passive segment hatching. Default is `[.4, .4, .4]`.
    removed_color : color, optional
        Color of removed segment hatching. Default is ``"r"`` if ``cmap`` is
        ``"viridis"``, and ``"b"`` otherwise.
    linewidth : int
        Width of lines. Default is ``2``.
    cmap : str, optional
        Name of a registered matplotlib colormap. If ``None`` (default), the
        current default colormap is used.
    """
    x_scale = x_scale.lower()

    if removed_color is not None:
        pass
    elif isinstance(cmap, str) and cmap == "viridis":
        removed_color = "r"
    else:
        removed_color = REMOVED_COLOR

    if signal_type is not None:
        pass
    elif "Sv" in transect:
        signal_type = "Sv"
    elif "signals" in transect:
        signal_type = "signals"
    else:
        raise ValueError(
            "No signal key found in transect. Available keys were: {}".format(
                transect.keys()
            )
        )

    turbulence_key = None
    for k in ("turbulence", "d_turbulence", "top", "d_top"):
        if k in transect:
            turbulence_key = k
            break

    bottom_key = None
    for k in ("bottom", "d_bottom", "bot", "d_bot"):
        if k in transect:
            bottom_key = k
            break

    surface_key = None
    for k in ("surface", "d_surface", "d_surf", "surf"):
        if k in transect:
            surface_key = k
            break

    if x_scale == "index":
        tt = np.arange(transect["timestamps"].shape[0])
        xlabel = "Ping index"
    elif x_scale == "timestamp":
        tt = transect["timestamps"]
        xlabel = "Timestamp (s)"
    elif x_scale == "time":
        tt = transect["timestamps"] - transect["timestamps"][0]
        xlabel = "Time (s)"
    else:
        raise ValueError("Unsupported x_scale: {}".format(x_scale))

    if signal_type == "Sv_masked":
        signal = copy.deepcopy(transect["Sv"])
        signal[~transect["mask"]] = np.nan
    else:
        signal = transect[signal_type]
    plt.pcolormesh(tt, transect["depths"], signal.T)
    if cmap is not None:
        plt.set_cmap(cmap)
    if surface_key is not None:
        plt.plot(tt, transect[surface_key], surface_color, linewidth=linewidth)
    if turbulence_key is not None:
        plt.plot(tt, transect[turbulence_key], turbulence_color, linewidth=linewidth)
    if bottom_key is not None:
        plt.plot(tt, transect[bottom_key], bottom_color, linewidth=linewidth)

    if show_regions:
        plot_indicator_hatch(
            transect.get("is_passive", []),
            xx=tt,
            ymin=transect["depths"][0],
            ymax=transect["depths"][-1],
            hatch="//",
            color=passive_color,
        )
        plot_indicator_hatch(
            transect.get("is_removed", []),
            xx=tt,
            ymin=transect["depths"][0],
            ymax=transect["depths"][-1],
            hatch=r"\\",
            color=removed_color,
        )
    if show_regions and "mask_patches" in transect:
        plot_mask_hatch(
            tt,
            transect["depths"],
            transect["mask_patches"],
            hatch=r"\\",
            color=removed_color,
            border=True,
        )

    plt.tick_params(reset=True, color=(0.2, 0.2, 0.2))
    plt.tick_params(labelsize=14)

    plt.xlabel(xlabel, fontsize=18)
    plt.ylabel("Depth (m)", fontsize=18)

    # Make sure y-axis is inverted (lowest depth at the top)
    ensure_axes_inverted(dir="y")


def plot_transect_predictions(transect, prediction, linewidth=1, cmap=None):
    r"""
    Plot the generated output for a transect against its ground truth data.

        - Ground truth data is shown in black, predictions in white.
        - Passive regions are hatched in / direction for ground truth, \ for
          prediciton.
        - Removed regions are hatched in \ direction for ground truth, / for
          prediction.

    Parameters
    ----------
    transect : dict
        Ground truth data for the transect.
    prediction : dict
        Predictions for the transect.
    linewidth : int
        Width of lines. Default is ``2``.
    cmap : str, optional
        Name of a registered matplotlib colormap. If ``None`` (default), the
        current default colormap is used.
    """
    plot_transect(
        transect,
        x_scale="index",
        turbulence_color="k",
        bottom_color="k",
        surface_color="k",
        passive_color="k",
        removed_color="k",
        linewidth=linewidth,
        cmap=cmap,
    )

    # Convert output into lines
    for shape_y in (
        prediction["p_is_above_turbulence"].shape[-1],
        prediction["p_is_below_bottom"].shape[-1],
    ):
        if not np.allclose(prediction["depths"].shape, shape_y):
            print("Shape mismatch: {} {}".format(prediction["depths"].shape, shape_y))
    surface_depths = prediction["depths"][
        utils.last_nonzero(prediction["p_is_above_surface"] > 0.5, -1)
    ]
    turbulence_depths = prediction["depths"][
        utils.last_nonzero(prediction["p_is_above_turbulence"] > 0.5, -1)
    ]
    bottom_depths = prediction["depths"][
        utils.first_nonzero(prediction["p_is_below_bottom"] > 0.5, -1)
    ]

    tt = np.linspace(0, len(transect["timestamps"]) - 1, len(turbulence_depths))
    plt.plot(tt, surface_depths, "w", linewidth=linewidth)
    plt.plot(tt, turbulence_depths, "w", linewidth=linewidth)
    plt.plot(tt, bottom_depths, "w", linewidth=linewidth)

    # Mark removed areas
    plot_indicator_hatch(
        prediction.get("p_is_passive", np.array([])) > 0.5,
        xx=tt,
        ymin=transect["depths"][0],
        ymax=transect["depths"][-1],
        hatch="\\\\",
        color="w",
    )
    plot_indicator_hatch(
        prediction.get("p_is_removed", np.array([])) > 0.5,
        xx=tt,
        ymin=transect["depths"][0],
        ymax=transect["depths"][-1],
        hatch="//",
        color="w",
    )
    if "p_is_patch" in prediction:
        plot_mask_hatch(
            tt,
            prediction["depths"],
            prediction["p_is_patch"] > 0.5,
            hatch="//",
            color="w",
            border=True,
        )
    # Make sure y-axis is inverted (lowest depth at the top)
    ensure_axes_inverted(dir="y")



"""
Convolutional layers.
"""

import functools
import itertools
import math
import numbers

import torch
from torch import nn
from torch.nn import functional as F

from .utils import _ntuple, same_to_padding


class Conv2dSame(nn.Conv2d):
    """
    2D Convolutions with same padding option.

    Same padding will only produce an output size which matches the input size
    if the kernel size is odd and the stride is 1.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding="same",
        dilation=1,
        **kwargs,
    ):
        if isinstance(padding, str) and padding == "same":
            padding = same_to_padding(kernel_size, stride, dilation, ndim=2)

        super(Conv2dSame, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            **kwargs,
        )


class PointwiseConv2d(nn.Conv2d):
    """
    2D Pointwise Convolution.
    """

    def __init__(self, in_channels, out_channels, **kwargs):
        if "kernel_size" in kwargs:
            raise ValueError("kernel_size must be 1 for a pointwise convolution.")
        super(PointwiseConv2d, self).__init__(
            in_channels, out_channels, kernel_size=1, **kwargs
        )


class DepthwiseConv2d(nn.Conv2d):
    """
    2D Depthwise Convolution.
    """

    def __init__(
        self, in_channels, kernel_size=3, stride=1, padding="same", dilation=1, **kwargs
    ):

        if "groups" in kwargs:
            raise ValueError(
                "Number of groups must equal number of input channels for a depthwise convolution."
            )

        if isinstance(padding, str) and padding == "same":
            padding = same_to_padding(kernel_size, stride, dilation, ndim=2)

        super(DepthwiseConv2d, self).__init__(
            in_channels,
            in_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=in_channels,
            **kwargs,
        )


class SeparableConv2d(nn.Module):
    """
    2D Depthwise Separable Convolution.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding="same",
        dilation=1,
        groups=1,
        **kwargs,
    ):
        super(SeparableConv2d, self).__init__()

        if isinstance(padding, str) and padding == "same":
            padding = same_to_padding(kernel_size, stride, dilation, ndim=2)

        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=in_channels,
            **kwargs,
        )

        self.pointwise = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            groups=groups,
            **kwargs,
        )

    def foward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)
        return x


class GaussianSmoothing(nn.Module):
    """
    Apply gaussian smoothing on a 1d, 2d or 3d tensor.

    Filtering is performed seperately for each channel in the input using a
    depthwise convolution.

    Parameters
    ----------
    channels : int or sequence
        Number of channels of the input tensors. Output will have this number
        of channels as well.
    kernel_size : int or sequence
        Size of the gaussian kernel.
    sigma : float or sequence
        Standard deviation of the gaussian kernel.
    padding : int or sequence or "same", optional
        Amount of padding to use, for each side of each dimension. If this is
        ``"same"`` (default) the amount of padding will be set automatically
        to ensure the size of the tensor is unchanged.
    pad_mode : str, optional
        Padding mode. See :meth:`torch.nn.functional.pad` for options. Default
        is ``"replicate"``.
    ndim : int, optional
        The number of dimensions of the data. Default value is 2 (spatial).

    Notes
    -----
    Based on
    https://discuss.pytorch.org/t/is-there-anyway-to-do-gaussian-filtering-for-an-image-2d-3d-in-pytorch/12351/10
    """

    def __init__(
        self, channels, kernel_size, sigma, padding="same", pad_mode="replicate", ndim=2
    ):
        super(GaussianSmoothing, self).__init__()

        # Ensure arguments are sequences of the correct length
        fntuple = _ntuple(ndim)
        kernel_size = fntuple(kernel_size)
        sigmas = fntuple(sigma)

        self.noop = all(s == 0 for s in sigmas)
        if self.noop:
            return

        # Handle padding arguments
        if isinstance(padding, str) and padding == "same":
            padding = same_to_padding(kernel_size, ndim=ndim)
            padding = tuple(
                itertools.chain.from_iterable(itertools.repeat(p, 2) for p in padding)
            )
        else:
            padding = _ntuple(ndim * 2)(padding)
        self.padding = padding
        self.pad_mode = pad_mode

        # The gaussian kernel is the product of the
        # gaussian function of each dimension.
        kernel = 1
        meshgrids = torch.meshgrid(
            [torch.arange(size, dtype=torch.float32) for size in kernel_size]
        )
        for size, std, mgrid in zip(kernel_size, sigmas, meshgrids):
            mean = (size - 1) / 2
            kernel *= torch.exp(-(((mgrid - mean) / std) ** 2) / 2) / (
                std * math.sqrt(2 * math.pi)
            )

        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel / torch.sum(kernel)

        # Reshape to depthwise convolutional weight
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = kernel.repeat(channels, *[1] * (kernel.dim() - 1))

        self.register_buffer("weight", kernel)
        self.groups = channels

        if ndim == 1:
            self.conv = F.conv1d
        elif ndim == 2:
            self.conv = F.conv2d
        elif ndim == 3:
            self.conv = F.conv3d
        else:
            raise ValueError(
                "Only 1, 2 and 3 dimensions are supported. Received {}.".format(ndim)
            )

    def forward(self, input):
        """
        Apply gaussian filter to input.

        Parameters
        ----------
        input : torch.Tensor
            Input to apply gaussian filter on.

        Returns
        -------
        filtered : torch.Tensor
            Filtered output, the same size as the input.
        """
        if self.noop:
            return input
        input = F.pad(input, self.padding, mode=self.pad_mode)
        return self.conv(input, weight=self.weight, groups=self.groups)



#!/usr/bin/env python

"""
Provides a command line interface for the training routine.

This is separated out from train.py so the documentation can be accessed without
having all the training dependencies installed.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import sys

from .. import __meta__
from . import formatters


def get_parser():
    """
    Build parser for training command line interface.

    Returns
    -------
    parser : argparse.ArgumentParser
        CLI argument parser for training.
    """
    prog = os.path.split(sys.argv[0])[1]
    if prog == "__main__.py" or prog == "__main__":
        prog = os.path.split(__file__)[1]
    parser = argparse.ArgumentParser(
        prog=prog,
        description="Echofilter model training",
        add_help=False,
    )

    # Actions
    group_action = parser.add_argument_group(
        "Actions",
        "These arguments specify special actions to perform. The main action"
        " of this program is supressed if any of these are given.",
    )
    group_action.add_argument(
        "-h",
        "--help",
        action="help",
        help="Show this help message and exit.",
    )
    group_action.add_argument(
        "--version",
        "-V",
        action="version",
        version="%(prog)s {version}".format(version=__meta__.version),
        help="Show program's version number and exit.",
    )

    # Data parameters
    group_data = parser.add_argument_group("Data parameters")
    group_data.add_argument(
        "--data-dir",
        type=str,
        default="/data/dsforce/surveyExports",
        metavar="DIR",
        help="path to root data directory",
    )
    group_data.add_argument(
        "--dataset",
        dest="dataset_name",
        type=str,
        default="mobile",
        help="which dataset to use",
    )
    group_data.add_argument(
        "--train-partition",
        type=str,
        default=None,
        help="which partition to train on (default depends on dataset)",
    )
    group_data.add_argument(
        "--val-partition",
        type=str,
        default=None,
        help="which partition to validate on (default depends on dataset)",
    )
    group_data.add_argument(
        "--shape",
        dest="sample_shape",
        nargs=2,
        type=int,
        default=(128, 512),
        help="input shape [W, H] (default: %(default)s)",
    )
    group_data.add_argument(
        "--crop-depth",
        type=float,
        default=None,
        help="depth, in metres, at which data should be truncated (default: %(default)s)",
    )
    group_data.add_argument(
        "--resume",
        default="",
        type=str,
        metavar="PATH",
        help='path to latest checkpoint (default: "%(default)s")',
    )
    group_data.add_argument(
        "--cold-restart",
        dest="restart",
        action="store_const",
        const="cold",
        default="",
        help="when resuming from a checkpoint, use this only for initial weights",
    )
    group_data.add_argument(
        "--warm-restart",
        dest="restart",
        action="store_const",
        const="warm",
        help="""
            when resuming from a checkpoint, use the existing weights and
            optimizer state but start a new LR schedule
        """,
    )
    group_data.add_argument(
        "--log",
        dest="log_name",
        default=None,
        type=str,
        help="output directory name (default: DATE_TIME)",
    )
    group_data.add_argument(
        "--log-append",
        dest="log_name_append",
        default=None,
        type=str,
        help="string to append to output directory name (default: HOSTNAME)",
    )

    # Model parameters
    group_model = parser.add_argument_group("Model parameters")
    group_model.add_argument(
        "--conditional",
        action="store_true",
        help=(
            "train a model conditioned on the direction the sounder is facing"
            " (in addition to an unconditional model)"
        ),
    )
    group_model.add_argument(
        "--nblock",
        "--num-blocks",
        dest="n_block",
        type=int,
        default=6,
        help="number of blocks down and up in the UNet (default: %(default)s)",
    )
    group_model.add_argument(
        "--latent-channels",
        type=int,
        default=32,
        help="number of initial/final latent channels to use in the model (default: %(default)s)",
    )
    group_model.add_argument(
        "--expansion-factor",
        type=float,
        default=1.0,
        help="expansion for number of channels as model becomes deeper"
        " (default: %(default)s, constant number of channels)",
    )
    group_model.add_argument(
        "--expand-only-on-down",
        action="store_true",
        help="only expand channels on dowsampling blocks",
    )
    group_model.add_argument(
        "--blocks-per-downsample",
        nargs="+",
        type=int,
        default=(2, 1),
        help="for each dim (time, depth), number of blocks between downsample"
        " steps (default: %(default)s)",
    )
    group_model.add_argument(
        "--blocks-before-first-downsample",
        nargs="+",
        type=int,
        default=(2, 1),
        help="for each dim (time, depth), number of blocks before first"
        " downsample step (default: %(default)s)",
    )
    group_model.add_argument(
        "--only-skip-connection-on-downsample",
        dest="always_include_skip_connection",
        action="store_false",
        help="only include skip connections when downsampling",
    )
    group_model.add_argument(
        "--deepest-inner",
        type=str,
        default="horizontal_block",
        help="layer to include at the deepest point of the UNet"
        ' (default: "horizontal_block"). Set to "identity" to disable.',
    )
    group_model.add_argument(
        "--intrablock-expansion",
        type=float,
        default=6.0,
        help="expansion within inverse residual blocks (default: %(default)s)",
    )
    group_model.add_argument(
        "--se-reduction",
        "--se",
        dest="se_reduction",
        type=float,
        default=4.0,
        help="reduction within squeeze-and-excite blocks (default: %(default)s)",
    )
    group_model.add_argument(
        "--downsampling-modes",
        nargs="+",
        type=str,
        default="max",
        help='for each downsampling step, the method to use (default: "%(default)s")',
    )
    group_model.add_argument(
        "--upsampling-modes",
        nargs="+",
        type=str,
        default="bilinear",
        help='for each upsampling step, the method to use (default: "%(default)s")',
    )
    group_model.add_argument(
        "--fused-conv",
        dest="depthwise_separable_conv",
        action="store_false",
        help="use fused instead of depthwise separable convolutions",
    )
    group_model.add_argument(
        "--no-residual",
        dest="residual",
        action="store_false",
        help="don't use residual blocks",
    )
    group_model.add_argument(
        "--actfn",
        type=str,
        default="InplaceReLU",
        help="activation function to use",
    )
    group_model.add_argument(
        "--kernel",
        dest="kernel_size",
        type=int,
        default=5,
        help="convolution kernel size (default: %(default)s)",
    )

    # Training methodology parameters
    group_training = parser.add_argument_group("Training parameters")
    group_training.add_argument(
        "--device",
        type=str,
        default="cuda",
        help='device to use (default: "%(default)s", using first gpu)',
    )
    group_training.add_argument(
        "--multigpu",
        action="store_true",
        help="train on multiple GPUs",
    )
    group_training.add_argument(
        "--no-amp",
        dest="use_mixed_precision",
        action="store_false",
        default=None,
        help="use fp32 instead of mixed precision (default: use mixed precision on gpu)",
    )
    group_training.add_argument(
        "--amp-opt",
        type=str,
        default="O1",
        help='optimizer level for apex automatic mixed precision (default: "%(default)s")',
    )
    group_training.add_argument(
        "-j",
        "--workers",
        dest="n_worker",
        type=int,
        default=8,
        metavar="N",
        help="number of data loading workers (default: %(default)s)",
    )
    group_training.add_argument(
        "-p",
        "--print-freq",
        type=int,
        default=50,
        help="print frequency (default: %(default)s)",
    )
    group_training.add_argument(
        "-b",
        "--batch-size",
        type=int,
        default=16,
        help="mini-batch size (default: %(default)s)",
    )
    group_training.add_argument(
        "--no-stratify",
        dest="stratify",
        action="store_false",
        help="disable stratified sampling; use fully random sampling instead",
    )
    group_training.add_argument(
        "--epochs",
        dest="n_epoch",
        type=int,
        default=20,
        help="number of total epochs to run (default: %(default)s)",
    )
    group_training.add_argument(
        "--seed",
        default=None,
        type=int,
        help="seed for initializing training.",
    )

    # Optimiser parameters
    group_optim = parser.add_argument_group("Optimizer parameters")
    group_optim.add_argument(
        "--optim",
        "--optimiser",
        "--optimizer",
        dest="optimizer",
        type=str,
        default="rangerva",
        help='optimizer name (default: "%(default)s")',
    )
    group_optim.add_argument(
        "--schedule",
        type=str,
        default="constant",
        help='LR schedule (default: "%(default)s")',
    )
    group_optim.add_argument(
        "--lr",
        "--learning-rate",
        dest="lr",
        type=float,
        default=0.1,
        metavar="LR",
        help="initial learning rate (default: %(default)s)",
    )
    group_optim.add_argument(
        "--momentum",
        type=float,
        default=0.9,
        help="momentum (default: %(default)s)",
    )
    group_optim.add_argument(
        "--base-momentum",
        type=float,
        default=None,
        help="base momentum; only used for OneCycle schedule (default: same as momentum)",
    )
    group_optim.add_argument(
        "--wd",
        "--weight-decay",
        dest="weight_decay",
        type=float,
        default=1e-5,
        help="weight decay (default: %(default)s)",
    )
    group_optim.add_argument(
        "--warmup-pct",
        type=float,
        default=0.2,
        help="fraction of training to spend warming up LR; only used for"
        " OneCycle MesaOneCycle schedules (default: %(default)s)",
    )
    group_optim.add_argument(
        "--warmdown-pct",
        type=float,
        default=0.7,
        help="fraction of training before warming down LR; only used for"
        " MesaOneCycle schedule (default: %(default)s)",
    )
    group_optim.add_argument(
        "--anneal-strategy",
        type=str,
        default="cos",
        help='annealing strategy; only used for OneCycle schedule (default: "%(default)s")',
    )
    group_optim.add_argument(
        "--overall-loss-weight",
        type=float,
        default=0.0,
        help="weighting for overall loss term (default: %(default)s)",
    )

    return parser


def _get_parser_sphinx():
    """
    Pre-format parser help for sphinx-argparse processing.
    """
    return formatters.format_parser_for_sphinx(get_parser())


def main(args=None):
    """
    Run command line interface for model training.
    """
    parser = get_parser()

    kwargs = vars(parser.parse_args(args))

    for k in ["blocks_per_downsample", "blocks_before_first_downsample"]:
        if len(kwargs[k]) == 1:
            kwargs[k] = kwargs[k][0]

    print("CLI arguments:")
    print(kwargs)
    print()

    # Use seaborn to set matplotlib plotting defaults
    import seaborn as sns

    sns.set()

    from ..train import train

    train(**kwargs)


if __name__ == "__main__":
    main()



"""
Path utilities.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys


def check_if_windows():
    """
    Check if the operating system is Windows.

    Returns
    -------
    bool
        Whether the OS is Windows.
    """
    return sys.platform.startswith("win")


def parse_files_in_folders(files_or_folders, source_dir, extension, recursive=True):
    """
    Walk through folders and find suitable files.

    Parameters
    ----------
    files_or_folders : iterable
        List of files and folders.
    source_dir : str
        Root directory within which elements of ``files_or_folders`` may
        be found.
    extension : str or Collection
        Extension (or list of extensions) which files within directories must
        bear to be included, without leading ``'.'``, for instance ``'.csv'``.
        Note that explicitly given files are always used.
    recursive : bool, optional
        Whether to walk through the tree of files in a subfolders of a
        directory input. If ``False``, only files in the folder itself and
        not its child folders will be included.

    Yields
    ------
    str
        Paths to explicitly given files and files within directories with
        extension ``extension``.
    """
    if extension is None or not isinstance(extension, str):
        extensions = extension
    else:
        extensions = {extension}
    if extensions is not None:
        extensions = {ext.lower() for ext in extensions}
    for path in files_or_folders:
        if check_if_windows():
            # Remove trailing slashes on Windows, because isdir("some\path\")
            # fails even though isdir("some\path") would succeed.
            path = path.rstrip("\\")

        if os.path.isfile(path) or os.path.isfile(os.path.join(source_dir, path)):
            yield path
            continue
        elif os.path.isdir(path):
            folder = path
        elif os.path.isdir(os.path.join(source_dir, path)):
            folder = os.path.join(source_dir, path)
        else:
            raise EnvironmentError("Missing file or directory: {}".format(path))

        if recursive:
            full_filenames = []
            for dirpath, _dirnames, fnames in os.walk(folder):
                for fname in fnames:
                    full_filenames.append(os.path.join(dirpath, fname))
        else:
            full_filenames = (
                os.path.join(folder, filename) for filename in os.listdir(folder)
            )

        for filename in full_filenames:
            if not os.path.isfile(filename):
                continue
            ext = os.path.splitext(filename)[1]
            if extensions is None or (len(ext) > 0 and ext[1:].lower() in extensions):
                yield filename


def determine_file_path(fname, source_dir):
    """
    Determine the path to use to an input file.

    Parameters
    ----------
    fname : str
        Path to an input file. Either an absolute path, or a path relative to
        to ``source_dir``, or a path relative to the working directory.
    source_dir : str
        Path to a directory where the file bearing name ``fname`` is expected to
        be located.

    Returns
    -------
    str
        Path to where file can be found, either absolute or relative.
    """
    # Check what the full path should be
    if os.path.isabs(fname) and os.path.isfile(fname):
        fname_full = fname
    elif os.path.isfile(os.path.join(source_dir, fname)):
        fname_full = os.path.join(source_dir, fname)
    elif os.path.isfile(fname):
        fname_full = fname
    else:
        raise EnvironmentError("Could not locate file {}".format(fname))

    return fname_full


def determine_destination(fname, fname_full, source_dir, output_dir):
    """
    Determine where destination should be placed for a file, preserving subtree paths.

    Parameters
    ----------
    fname : str
        Original input path.
    fname_full : str
        Path to file, either absolute or relative; possibly containing
        ``source_dir``.
    source_dir : str
        Path to a directory where the file bearing name ``fname`` is expected to
        be located.
    output_dir : str
        Path to root output directory.

    Returns
    -------
    str
        Path to where file can be found, either absolute or relative.
    """
    # Determine where destination should be placed
    if output_dir is None or output_dir == "":
        return fname_full
    if os.path.isabs(fname):
        return os.path.join(output_dir, os.path.split(fname)[1])
    full_source_dir = os.path.join(os.path.abspath(source_dir), "")
    if os.path.abspath(fname).startswith(full_source_dir):
        return os.path.join(output_dir, os.path.abspath(fname)[len(full_source_dir) :])
    return os.path.join(output_dir, fname)



"""
Blocks of modules.
"""

import torch
from torch import nn

from .activations import str2actfnfactory
from .conv import Conv2dSame, DepthwiseConv2d, PointwiseConv2d
from .pathing import ResidualConnect
from .utils import _pair

__all__ = ["MBConv", "SqueezeExcite"]


class SqueezeExcite(nn.Module):
    """
    Squeeze and excitation block.

    See https://arxiv.org/abs/1709.01507

    Parameters
    ----------
    in_channels : int
        Number of input (and output) channels.
    reduction : int or float, optional
        Compression factor for the number of channels in the squeeze and
        excitation attention module. Default is ``4``.
    actfn : str or callable, optional
        An activation class or similar generator. Default is an inplace
        ReLU activation. If this is a string, it is mapped to a generator with
        ``activations.str2actfnfactory``.
    """

    def __init__(
        self,
        in_channels,
        reduction=4,
        actfn="InplaceReLU",
    ):
        super(SqueezeExcite, self).__init__()

        actfn_factory = actfn
        if isinstance(actfn, str):
            actfn_factory = str2actfnfactory(actfn)

        reduced_chns = int(max(1, round(in_channels / reduction)))

        layers = [
            nn.AdaptiveAvgPool2d(1),
            PointwiseConv2d(in_channels, reduced_chns),
            actfn_factory(),
            PointwiseConv2d(reduced_chns, in_channels),
            nn.Sigmoid(),
        ]
        self.layers = nn.Sequential(*layers)

    def forward(self, input):
        return input * self.layers(input)


class MBConv(nn.Module):
    """
    MobileNet style inverted residual block.

    See https://arxiv.org/abs/1905.11946 and https://arxiv.org/abs/1905.02244.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int, optional
        Number of output channels. Default is to match ``in_channels``.
    expansion : int or float, optional
        Exansion factor for the inverted-residual bottleneck. Default is ``6``.
    se_reduction : int, optional
        Reduction factor for squeeze-and-excite block. Default is ``4``. Set
        to ``None`` or ``0`` to disable squeeze-and-excitation.
    fused : bool, optional
        If ``True``, the pointwise and depthwise convolution are fused together
        into a single regular convolution. Default is ``False`` (a depthwise
        separable convolution).
    residual : bool, optional
        If ``True``, the block is residual with a skip-through connection.
        Default is ``True``.
    actfn : str or callable, optional
        An activation class or similar generator. Default is an inplace
        ReLU activation. If this is a string, it is mapped to a generator with
        ``activations.str2actfnfactory``.
    bias : bool, optional
        If ``True``, the main convolution has a bias term. Default is ``False``.
        Note that the pointwise convolutions never have bias terms.
    **conv_args
        Additional arguments, such as kernel_size, stride, and padding, which
        will be passed to the convolution module.
    """

    def __init__(
        self,
        in_channels,
        out_channels=None,
        expansion=6,
        se_reduction=4,
        fused=False,
        residual=True,
        actfn="InplaceReLU",
        bias=False,
        **conv_args,
    ):
        super(MBConv, self).__init__()

        if out_channels is None:
            out_channels = in_channels

        self.residual = residual
        self.fused = fused

        actfn_factory = actfn
        if isinstance(actfn, str):
            actfn_factory = str2actfnfactory(actfn)

        expanded_chns = int(round(in_channels * expansion))

        if expansion == 1 or fused:
            self.expansion_conv = nn.Identity()
        else:
            self.expansion_conv = nn.Sequential(
                PointwiseConv2d(in_channels, expanded_chns, bias=False),
                nn.BatchNorm2d(expanded_chns),
                actfn_factory(),
            )
        if fused:
            conv = Conv2dSame(in_channels, expanded_chns, bias=bias, **conv_args)
        else:
            conv = DepthwiseConv2d(expanded_chns, bias=bias, **conv_args)
        self.conv = nn.Sequential(conv, nn.BatchNorm2d(expanded_chns), actfn_factory())

        if se_reduction:
            self.se = SqueezeExcite(expanded_chns, reduction=se_reduction)
        else:
            self.se = nn.Identity()

        self.contraction_conv = nn.Sequential(
            PointwiseConv2d(expanded_chns, out_channels, bias=False),
            nn.BatchNorm2d(out_channels),
        )

        if residual:
            if any([k > 1 for k in _pair(conv_args.get("stride", 1))]):
                self.skip_pool = nn.AvgPool2d(conv_args["stride"])
            else:
                self.skip_pool = nn.Identity()
            self.connector = ResidualConnect(in_channels, out_channels)

    def forward(self, input):
        x = self.expansion_conv(input)
        x = self.conv(x)
        x = self.se(x)
        x = self.contraction_conv(x)
        if self.residual:
            x = self.connector(x, self.skip_pool(input))
        return x

    def extra_repr(self):
        return "residual={residual}, fused={fused}".format(
            residual=self.residual,
            fused=self.fused,
        )



"""
Convert echograms into Pytorch dataset.

Tools for converting a dataset of echograms (transects) into a Pytorch dataset
and sampling from it.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import random
import warnings

import numpy as np
import torch.utils.data

from ..raw import shardloader
from . import utils


class TransectDataset(torch.utils.data.Dataset):
    """
    Load a collection of transects as a PyTorch dataset.

    Parameters
    ----------
    transect_paths : list
        Absolute paths to transects.
    window_len : int
        Width (number of timestamps) to load. Default is ``128``.
    p_scale_window : float, optional
        Probability of rescaling window. Default is ``0``, which results in no
        randomization of the window widths.
    window_sf : float, optional
        Maximum window scale factor. Scale factors will be log-uniformly
        sampled in the range ``1/window_sf`` to ``window_sf``. Default is ``2``.
    num_windows_per_transect : int
        Number of windows to extract for each transect. Start indices for
        the windows will be equally spaced across the total width of the
        transect. If this is ``0``, the number of windows will be inferred
        automatically based on ``window_len`` and the total width of the
        transect, resulting in a different number of windows for each
        transect. Default is ``0``.
    use_dynamic_offsets : bool
        Whether starting indices for each window should be randomly offset.
        Set to ``True`` for training and ``False`` for testing. Default is
        ``True``.
    crop_depth : float
        Maximum depth to include, in metres. Deeper data will be cropped
        away. Default is ``None``.
    transform : callable
        Operations to perform to the dictionary containing a single sample.
        These are performed before generating the turbulence/bottom/overall
        mask. Default is ``None``.
    remove_nearfield : bool, optional
        Whether to remove turbulence and bottom lines affected by nearfield
        removal. If ``True`` (default), targets for the line near to the
        sounder (bottom if upward facing, turbulence otherwise) which are
        closer than or equal to a distance of ``nearfield_distance`` become
        reduced to ``nearfield_visible_dist``.
    nearfield_distance : float, optional
        Nearfield distance in metres. Regions closer than the nearfield
        may have been masked out from the dataset, but their effect will
        be removed from the targets if ``remove_nearfield=True``.
        Default is ``1.7``.
    nearfield_visible_dist : float, optional
        The distance at which the effect of being to close to the sounder
        is obvious to the naked eye, and hence the distance which nearfield
        will be mapped to if ``remove_nearfield=True``. Default is ``0.0``.
    remove_offset_turbulence : float, optional
        Line offset built in to the turbulence line. If given, this will be
        removed from the samples within the dataset. Default is ``0``.
    remove_offset_bottom : float, optional
        Line offset built in to the bottom line. If given, this will be
        removed from the samples within the dataset. Default is ``0``.
    """

    def __init__(
        self,
        transect_paths,
        window_len=128,
        p_scale_window=0,
        window_sf=2,
        num_windows_per_transect=0,
        use_dynamic_offsets=True,
        crop_depth=None,
        transform=None,
        remove_nearfield=True,
        nearfield_distance=1.7,
        nearfield_visible_dist=0.0,
        remove_offset_turbulence=0,
        remove_offset_bottom=0,
    ):
        super(TransectDataset, self).__init__()
        self.transect_paths = transect_paths
        self.window_len = window_len
        self.p_scale_window = p_scale_window
        self.max_window_sf = window_sf if p_scale_window else 1
        self.num_windows = num_windows_per_transect
        self.use_dynamic_offsets = use_dynamic_offsets
        self.crop_depth = crop_depth
        self.transform = transform
        self.remove_nearfield = remove_nearfield
        self.nearfield_distance = nearfield_distance
        self.nearfield_visible_dist = nearfield_visible_dist
        self.remove_offset_turbulence = remove_offset_turbulence
        self.remove_offset_bottom = remove_offset_bottom
        self.initialise_datapoints()

    def initialise_datapoints(self):
        """
        Parse ``transect_paths`` to generate sampling windows for each transect.

        Manually calling this method will resample the transect offsets and
        widths if they were randomly generated.
        """
        self.datapoints = []

        for transect_path in self.transect_paths:
            # Check how many segments the transect was divided into
            segments_meta_fname = os.path.join(transect_path, "n_segment.txt")
            if not os.path.isfile(segments_meta_fname):
                # Silently skip missing transects
                continue
            with open(segments_meta_fname, "r") as f:
                n_segment = int(f.readline().strip())

            # For each segment, specify some samples over its duration
            for i_segment in range(n_segment):
                seg_path = os.path.join(transect_path, str(i_segment))
                # Lookup the number of rows in the transect
                # Load the sharding metadata
                with open(os.path.join(seg_path, "shard_size.txt"), "r") as f:
                    n_timestamps, shard_len = f.readline().strip().split(",")
                    n_timestamps = int(n_timestamps)
                # Generate an array for window centers within the transect
                # - if this is for training, we want to randomise the offsets
                # - if this is for validation, we want stable windows
                num_windows = self.num_windows
                if self.num_windows is None or self.num_windows == 0:
                    # Load enough windows to include all datapoints
                    num_windows = int(np.ceil(n_timestamps / self.window_len))
                centers = np.linspace(0, n_timestamps, num_windows + 1)[:num_windows]
                if len(centers) > 1:
                    max_dy_offset = centers[1] - centers[0]
                else:
                    max_dy_offset = n_timestamps
                if self.use_dynamic_offsets:
                    centers += random.random() * max_dy_offset
                else:
                    centers += max_dy_offset / 2
                centers = np.round(centers)
                # Generate a width for each window, and add each
                # (transect, center, width) to the list for this epoch
                for center_idx in centers:
                    cur_win_len = self.window_len
                    if self.p_scale_window and random.random() < self.p_scale_window:
                        sf = np.exp(np.log(self.max_window_sf) * random.uniform(-1, 1))
                        cur_win_len *= sf
                    cur_win_len = int(np.round(cur_win_len))
                    self.datapoints.append((seg_path, int(center_idx), cur_win_len))

    def __getitem__(self, index):
        transect_pth, center_idx, win_len = self.datapoints[index]
        # Load data from shards
        sample = shardloader.load_transect_from_shards_abs(
            transect_pth,
            center_idx - int(win_len / 2),
            center_idx - int(win_len / 2) + win_len,
            pad_mode="reflect",
        )

        # Fixup sample
        sample = fixup_dataset_sample(
            sample,
            remove_nearfield=self.remove_nearfield,
            nearfield_distance=self.nearfield_distance,
            nearfield_visible_dist=self.nearfield_visible_dist,
            remove_offset_turbulence=self.remove_offset_turbulence,
            remove_offset_bottom=self.remove_offset_bottom,
            crop_depth=self.crop_depth,
            transform=self.transform,
        )

        # Rename keys for depth of lines
        sample["d_turbulence"] = sample.pop("turbulence")
        sample["d_bottom"] = sample.pop("bottom")
        sample["d_surface"] = sample.pop("surface")
        sample["d_turbulence-original"] = sample.pop("turbulence-original")
        sample["d_bottom-original"] = sample.pop("bottom-original")
        sample["signals"] = sample.pop("Sv")
        if sample["depths"][-1] < sample["depths"][0]:
            # Found some upward-facing data that needs to be reflected
            for k in ["depths", "signals", "mask"]:
                sample[k] = np.flip(sample[k], -1).copy()

        # Convert mask patches into floating point arrays
        for suffix in ("", "-original", "-ntob"):
            sample["mask_patches" + suffix] = sample["mask_patches" + suffix].astype(
                np.float32
            )
        sample["mask"] = sample["mask"].astype(np.float32)

        # Add relative depth for lines
        depth_range = abs(sample["depths"][-1] - sample["depths"][0])
        for key in [
            "d_turbulence",
            "d_bottom",
            "d_surface",
            "d_turbulence-original",
            "d_bottom-original",
        ]:
            sample["r" + key[1:]] = sample[key] / depth_range

        # Ensure everything is float32 datatype
        for key in sample:
            sample[key] = sample[key].astype(np.float32)
            # Ensure no entries are invalid numbers
            nan_count = np.sum(np.isnan(sample[key]))
            if nan_count > 0:
                print("WARNING: NaN present in {}".format(key))
                sample[key][np.isnan(sample[key])] = 0
            inf_count = np.sum(np.isinf(sample[key]))
            if inf_count > 0:
                print("WARNING: inf present in {}".format(key))
                sample[key][np.isinf(sample[key])] = 0
            inf_count = np.sum(~np.isfinite(sample[key]))
            if inf_count > 0:
                print("WARNING: non-finite numbers present in {}".format(key))
                sample[key][~np.isfinite(sample[key])] = 0

        input = np.expand_dims(sample["signals"], 0)
        return input, sample

    def __len__(self):
        return len(self.datapoints)


class ConcatDataset(torch.utils.data.ConcatDataset):
    """
    Dataset as a concatenation of multiple TransectDatasets.

    This class is useful to assemble different existing datasets.

    Parameters
    ----------
    datasets : sequence
        List of datasets to be concatenated.

    Notes
    -----
    A subclass of :class:`torch.utils.data.ConcatDataset` which supports the
    ``initialise_datapoints`` method.
    """

    def initialise_datapoints(self):
        for dataset in self.datasets:
            dataset.initialise_datapoints()


class StratifiedRandomSampler(torch.utils.data.Sampler):
    """
    Sample elements randomly without repetition, stratified across datasets.

    Parameters
    ----------
    data_source : torch.utils.data.ConcatDataset
        Dataset to sample from. Must possess a ``cumulative_sizes`` attribute.
    """

    def __init__(self, data_source):
        self.data_source = data_source

    @property
    def num_samples(self):
        # dataset size might change at runtime
        return len(self.data_source)

    def __iter__(self):

        n_sample = len(self.data_source)
        n_dataset = len(self.data_source.cumulative_sizes)

        perms = []
        lower = 0
        for upper in self.data_source.cumulative_sizes:
            p = list(range(lower, upper))
            random.shuffle(p)
            perms.append(p)
            lower = upper

        dataset_sizes = np.array([len(p) for p in perms])
        target_fraction = dataset_sizes / n_sample
        sub_tallies = np.zeros(n_dataset, dtype=int)
        cur_fraction = np.zeros(n_dataset)

        indices = []
        while len(indices) < n_sample:
            i = np.argmax(target_fraction - cur_fraction)
            indices.append(perms[i].pop())
            sub_tallies[i] += 1
            cur_fraction = sub_tallies / len(indices)

        return iter(indices)

    def __len__(self):
        return self.num_samples


def fixup_dataset_sample(
    sample,
    remove_nearfield=True,
    nearfield_distance=1.7,
    nearfield_visible_dist=0.0,
    remove_offset_turbulence=0.0,
    remove_offset_bottom=0.0,
    crop_depth=None,
    transform=None,
):
    """
    Handle a dataset transect sample.

    Parameters
    ----------
    sample : dict
        Transect dictionary.
    remove_nearfield : bool, default=True
        Whether to remove turbulence and bottom lines affected by nearfield
        removal. If ``True`` (default), targets for the line near to the
        sounder (bottom if upward facing, turbulence otherwise) which are
        closer than or equal to a distance of ``nearfield_distance`` become
        reduced to ``nearfield_visible_dist``.
    nearfield_distance : float, default=1.7
        Nearfield distance in metres. Regions closer than the nearfield
        may have been masked out from the dataset, but their effect will
        be removed from the targets if ``remove_nearfield=True``.
    nearfield_visible_dist : float, default=0
        The distance at which the effect of being to close to the sounder
        is obvious to the naked eye, and hence the distance which nearfield
        will be mapped to if ``remove_nearfield=True``.
    remove_offset_turbulence : float, default=0
        Line offset built in to the turbulence line. If given, this will be
        removed from the samples within the dataset.
    remove_offset_bottom : float, default=0
        Line offset built in to the bottom line. If given, this will be
        removed from the samples within the dataset.
    crop_depth : float
        Maximum depth to include, in metres. Deeper data will be cropped
        away. Default is ``None``.
    transform : callable, optional
        Operations to perform to the dictionary containing a single sample.
        These are performed before generating the turbulence/bottom/overall
        mask.

    Returns
    -------
    dict
        Like ``sample``, but contents fixed.
    """
    # Rename Sv to signals if necessary
    if "signals" not in sample and "Sv" in sample:
        sample["signals"] = sample.pop("Sv")

    if sample["depths"][-1] < sample["depths"][0]:
        # Found some upward-facing data that needs to be reflected
        for k in ["depths", "signals", "mask"]:
            sample[k] = np.flip(sample[k], -1).copy()

    sample["depths"] = sample["depths"].astype(np.float32)

    # Fix any broken surface lines (these were generated automatically
    # by Echoview and not adjusted by human annotator, so are not
    # guaranteed to be sane).
    # Note any locations where the labels are not sane. These can be
    # removed from the training loss.
    sample["is_bad_labels"] = sample["surface"] >= sample["bottom"]
    # Surface line can not be below turbulence line
    sample["surface"] = np.minimum(sample["surface"], sample["turbulence"])
    # Ensure the bottom line is always below the surface line as well
    sample["bottom"] = np.maximum(sample["bottom"], sample["surface"] + 0.02)

    if sample["is_upward_facing"]:
        min_top_depth = np.min(sample["depths"])
        max_bot_depth = np.max(sample["depths"]) - nearfield_visible_dist
    else:
        min_top_depth = 0.966
        max_bot_depth = np.max(sample["depths"])

    if remove_nearfield:
        depth_intv = np.abs(sample["depths"][-1] - sample["depths"][-2])
        if sample["is_upward_facing"]:
            nearfield_threshold = (
                np.max(sample["depths"]) - depth_intv * 2.5 - nearfield_distance
            )
            was_in_nearfield = sample["bottom"] >= nearfield_threshold
            sample["bottom"][was_in_nearfield] = max_bot_depth
            was_in_nearfield_og = sample["bottom-original"] >= nearfield_threshold
            sample["bottom-original"][was_in_nearfield_og] = max_bot_depth
            # Extend/contract mask_patches where necessary
            idx_search = utils.last_nonzero(sample["depths"] < nearfield_threshold)
            idx_fillto = utils.first_nonzero(sample["depths"] > max_bot_depth)
            is_close_patch = np.any(
                sample["mask_patches"][:, idx_search:idx_fillto], -1
            )
            sample["mask_patches"][is_close_patch, idx_search:idx_fillto] = 1
            is_close_patch_og = sample["mask_patches-original"][:, idx_search] > 0
            is_close_patch_og = np.expand_dims(is_close_patch_og, -1)
            sample["mask_patches-original"][:, idx_search:] = is_close_patch_og
            sample["mask_patches-ntob"][:, idx_search:] = is_close_patch_og
        else:
            was_in_nearfield = sample["turbulence"] < nearfield_distance + depth_intv
            sample["turbulence"][was_in_nearfield] = min_top_depth
            was_in_nearfield_og = np.zeros_like(sample["is_removed"], dtype="bool")
            # Extend/contract mask_patches where necessary
            idx_search = utils.first_nonzero(
                sample["depths"] > nearfield_distance, invalid_val=0
            )
            idx_fillfr = utils.last_nonzero(
                sample["depths"] < min_top_depth, invalid_val=-1
            )
            is_close_patch = np.any(
                sample["mask_patches"][:, idx_fillfr + 1 : idx_search + 1], -1
            )
            sample["mask_patches"][is_close_patch, idx_fillfr + 1 : idx_search] = 1
            is_close_patch = np.any(
                sample["mask_patches-ntob"][:, idx_fillfr + 1 : idx_search + 1], -1
            )
            sample["mask_patches-ntob"][is_close_patch, idx_fillfr + 1 : idx_search] = 1
            is_close_patch_og = sample["mask_patches-original"][:, idx_search] > 0
            is_close_patch_og = np.expand_dims(is_close_patch_og, -1)
            sample["mask_patches-original"][:, :idx_search] = is_close_patch_og
    else:
        was_in_nearfield = np.zeros_like(sample["is_removed"], dtype="bool")
        was_in_nearfield_og = np.zeros_like(sample["is_removed"], dtype="bool")

    if remove_offset_turbulence:
        # Check the mask beforehand
        _ddepths = np.broadcast_to(sample["depths"], sample["signals"].shape)
        _in_mask = _ddepths < np.expand_dims(sample["turbulence"], -1)
        _in_mask_og = _ddepths < np.expand_dims(sample["turbulence-original"], -1)
        # Shift lines up higher (less deep)
        sample["turbulence"][
            (~was_in_nearfield) | sample["is_upward_facing"]
        ] -= remove_offset_turbulence
        sample["turbulence-original"][
            (~was_in_nearfield_og) | sample["is_upward_facing"]
        ] -= remove_offset_turbulence
        # Extend mask_patches where necessary
        _fx_mask = _ddepths < np.expand_dims(sample["turbulence"], -1)
        _df_mask = _in_mask * (_in_mask ^ _fx_mask)
        is_close_patch = np.any(_df_mask[:, :-1] * sample["mask_patches"][:, 1:], -1)
        sample["mask_patches"][is_close_patch, :] += _df_mask[is_close_patch, :]
        # ... and extend ntob mask patches too
        sample["mask_patches-ntob"][is_close_patch, :] += _df_mask[is_close_patch, :]
        # ... and extend og mask patches too
        _fx_mask_og = _ddepths < np.expand_dims(sample["turbulence-original"], -1)
        _df_mask = _in_mask_og * (_in_mask_og ^ _fx_mask_og)
        is_close_patch = np.any(
            _df_mask[:, :-1] * sample["mask_patches-original"][:, 1:], -1
        )
        sample["mask_patches-original"][is_close_patch, :] += _df_mask[
            is_close_patch, :
        ]

    if remove_offset_bottom:
        # Check the mask beforehand
        _ddepths = np.broadcast_to(sample["depths"], sample["signals"].shape)
        _in_mask = _ddepths > np.expand_dims(sample["bottom"], -1)
        _in_mask_og = _ddepths > np.expand_dims(sample["bottom-original"], -1)
        # Shift lines down lower (more deep)
        sample["bottom"][
            (~was_in_nearfield) | ~sample["is_upward_facing"]
        ] += remove_offset_bottom
        sample["bottom-original"][
            (~was_in_nearfield_og) | sample["is_upward_facing"]
        ] += remove_offset_bottom
        # Extend mask_patches where necessary
        _fx_mask = _ddepths > np.expand_dims(sample["bottom"], -1)
        _df_mask = _in_mask * (_in_mask ^ _fx_mask)
        is_close_patch = np.any(_df_mask[:, 1:] * sample["mask_patches"][:, :-1], -1)
        sample["mask_patches"][is_close_patch, :] += _df_mask[is_close_patch, :]
        # ... and extend og mask patches too
        _fx_mask_og = _ddepths > np.expand_dims(sample["bottom-original"], -1)
        _df_mask = _in_mask_og * (_in_mask_og ^ _fx_mask_og)
        is_close_patch = np.any(
            _df_mask[:, 1:] * sample["mask_patches-original"][:, :-1], -1
        )
        sample["mask_patches-original"][is_close_patch, :] += _df_mask[
            is_close_patch, :
        ]
        # ... and extend ntob mask patches too
        sample["mask_patches-ntob"][is_close_patch, :] += _df_mask[is_close_patch, :]

    if remove_offset_turbulence or remove_offset_bottom:
        # Change any 2s in the mask to be 1s
        for suffix in ("", "-original", "-ntob"):
            sample["mask_patches" + suffix] = sample["mask_patches" + suffix] > 0.5

    # Ensure all line values are finite
    for key in [
        "surface",
        "turbulence",
        "turbulence-original",
        "bottom",
        "bottom-original",
    ]:
        where_invalid = ~np.isfinite(sample[key])
        if np.sum(where_invalid) > 0 and np.sum(where_invalid) < len(sample[key]):
            sample[key][where_invalid] = np.interp(
                sample["timestamps"][where_invalid],
                sample["timestamps"][~where_invalid],
                sample[key][~where_invalid],
            )

    # Apply depth crop
    if crop_depth is not None:
        depth_crop_mask = sample["depths"] <= crop_depth
        sample["depths"] = sample["depths"][depth_crop_mask]
        sample["signals"] = sample["signals"][:, depth_crop_mask]

    if transform is not None:
        sample = transform(sample)

    # Convert lines to masks
    ddepths = np.broadcast_to(sample["depths"], sample["signals"].shape)
    for suffix in ["", "-original"]:
        sample["mask_turbulence" + suffix] = ddepths < np.expand_dims(
            sample["turbulence" + suffix], -1
        )
        sample["mask_bottom" + suffix] = ddepths > np.expand_dims(
            sample["bottom" + suffix], -1
        )
    sample["mask_surface"] = ddepths < np.expand_dims(sample["surface"], -1)

    # Ensure is_surrogate_surface excludes passive data
    sample["is_surrogate_surface"][sample["is_passive"] > 0.5] = 0

    # Create mask corresponding to the aggregate of all elements we need
    # masked in/out
    sample["mask"] = np.ones_like(sample["signals"], dtype="bool")
    sample["mask"][sample["is_passive"] > 0.5] = 0
    sample["mask"][sample["is_removed"] > 0.5] = 0
    sample["mask"][sample["mask_turbulence"] > 0.5] = 0
    sample["mask"][sample["mask_bottom"] > 0.5] = 0
    sample["mask"][sample["mask_patches"] > 0.5] = 0

    # Determine the boundary index for depths
    for sfx in {"turbulence", "turbulence-original", "surface"}:
        # Ties are broken to the smaller index
        sample["index_" + sfx] = np.searchsorted(
            sample["depths"], sample[sfx], side="left"
        )
        # It is possible for the turbulence line to be above the field of
        # view, and impossible for the turbulence line to below
        sample["index_" + sfx] = np.maximum(
            0, np.minimum(len(sample["depths"]) - 1, sample["index_" + sfx])
        )
    for sfx in {"bottom", "bottom-original"}:
        # Ties are broken to the larger index
        sample["index_" + sfx] = np.searchsorted(
            sample["depths"], sample[sfx], side="right"
        )
        # It is possible for the bottom line to be below the field of view,
        # and impossible for the bottom line to above
        sample["index_" + sfx] -= 1
        sample["index_" + sfx] = np.maximum(
            0, np.minimum(len(sample["depths"]) - 1, sample["index_" + sfx])
        )

    return sample



"""
Interacting with the list of available checkpoints.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import pickle
from collections import OrderedDict

import appdirs
import yaml

from . import style

PACKAGE_DIR = os.path.dirname(os.path.dirname(__file__))
REPO_DIR = os.path.dirname(PACKAGE_DIR)
CHECKPOINT_FILE = os.path.join(PACKAGE_DIR, "checkpoints.yaml")
CHECKPOINT_FILE_ALT = os.path.join(REPO_DIR, "checkpoints.yaml")
CHECKPOINT_EXT = ".pt"


def get_checkpoint_list():
    """
    List the currently available checkpoints, as stored in a local file.

    Returns
    -------
    checkpoints : OrderedDict
        Dictionary with a key for each checkpoint. Each key maps to a dictionary
        whose elements describe the checkpoint.
    """
    checkpoint_file_use = None
    if os.path.isfile(CHECKPOINT_FILE):
        checkpoint_file_use = CHECKPOINT_FILE
    elif os.path.isfile(CHECKPOINT_FILE_ALT):
        checkpoint_file_use = CHECKPOINT_FILE_ALT
    else:
        raise EnvironmentError(f"No such file: '{CHECKPOINT_FILE}'")
    with open(checkpoint_file_use, "r") as hf:
        checkpoints = OrderedDict(yaml.safe_load(hf)["checkpoints"])
    return checkpoints


def get_default_checkpoint():
    """
    Get the name of the current default checkpoint.

    Returns
    -------
    checkpoint_name : str
        Name of current checkpoint.
    """
    return next(iter(get_checkpoint_list()))


def cannonise_checkpoint_name(name):
    """
    Cannonises checkpoint name by removing extension.

    Parameters
    ----------
    name : str
        Name of checkpoint, possibly including extension.

    Returns
    -------
    name : str
        Name of checkpoint, with extension removed it matches a possible
        checkpoint file extension.
    """
    for possible_ext in [
        ".ckpt.pth.tar",
        ".checkpoint.pth.tar",
        ".pth.tar",
        ".ckpt.tar",
        ".checkpoint.tar",
        CHECKPOINT_EXT,
        ".pt",
        ".pth",
        ".ckpt",
        ".tar",
    ]:
        if name.lower().endswith(possible_ext):
            name = name[: -len(possible_ext)]
            return name
    return name


class ListCheckpoints(argparse.Action):
    def __call__(self, parser, namespace, values, option_string):
        print("Currently available model checkpoints:")
        for checkpoint in get_checkpoint_list():
            if checkpoint == get_default_checkpoint():
                print("  * " + style.progress_fmt(checkpoint))
            else:
                print("    " + checkpoint)
        parser.exit()  # exits the program with no more arg parsing and checking


def get_default_cache_dir():
    """Determine the default cache directory."""
    return appdirs.user_cache_dir("echofilter", "DeepSense")


def download_checkpoint(checkpoint_name, cache_dir=None, verbose=1):
    """
    Download a checkpoint if it isn't already cached.

    Parameters
    ----------
    checkpoint_name : str
        Name of checkpoint to download.
    cache_dir : str or None, optional
        Path to local cache directory. If ``None`` (default), an OS-appropriate
        application-specific default cache directory is used.
    verbose : int, optional
        Verbosity level. Default is ``1``. Set to ``0`` to disable print
        statements.

    Returns
    -------
    str
        Path to downloaded checkpoint file.
    """
    if cache_dir is None:
        cache_dir = get_default_cache_dir()

    checkpoint_name = cannonise_checkpoint_name(checkpoint_name)
    destination = os.path.join(cache_dir, checkpoint_name + CHECKPOINT_EXT)

    if os.path.exists(destination):
        return destination

    # Import packages needed for downloading files
    import urllib

    import requests
    from torchvision.datasets.utils import download_file_from_google_drive, download_url

    os.makedirs(cache_dir, exist_ok=True)

    sources = get_checkpoint_list()[checkpoint_name]
    success = False
    for key, url_or_id in sources.items():
        if key == "gdrive":
            if verbose >= 1:
                print(
                    "Downloading checkpoint {} from GDrive...".format(checkpoint_name)
                )
            try:
                download_file_from_google_drive(
                    url_or_id,
                    os.path.dirname(destination),
                    filename=os.path.basename(destination),
                )
                success = True
                continue
            except pickle.UnpicklingError:
                if verbose >= 1:
                    print(
                        style.error_fmt(
                            "\nCould not download checkpoint {} from GDrive!".format(
                                checkpoint_name
                            )
                        )
                    )
            except (requests.exceptions.ConnectionError, urllib.error.URLError):
                msg = "Could not connect to Google Drive. Please check your Internet connection."
                with style.error_message(msg) as msg:
                    raise EnvironmentError(msg)
        else:
            if verbose >= 1:
                print(
                    "Downloading checkpoint {} from {}...".format(
                        checkpoint_name, url_or_id
                    )
                )
            try:
                download_url(url_or_id, cache_dir, filename=checkpoint_name)
                success = True
                continue
            except pickle.UnpicklingError:
                if verbose >= 1:
                    print(
                        style.error_fmt(
                            "\nCould not download checkpoint {} from {}".format(
                                checkpoint_name, url_or_id
                            )
                        )
                    )
            except (requests.exceptions.ConnectionError, urllib.error.URLError):
                msg = "Could not connect to file server to download {}. Please check your Internet connection.".format(
                    url_or_id
                )
                with style.error_message(msg) as msg:
                    raise EnvironmentError(msg)

    if not success:
        msg = "Unable to download {} from {}".format(checkpoint_name, sources)
        with style.error_message(msg) as msg:
            raise OSError(msg)

    if verbose >= 1:
        print("Downloaded checkpoint to {}".format(destination))

    return destination


def load_checkpoint(
    ckpt_name=None, cache_dir=None, device="cpu", return_name=False, verbose=1
):
    """
    Load a checkpoint, either from absolute path or the cache.

    Parameters
    ----------
    checkpoint_name : str or None, optional
        Path to checkpoint file, or name of checkpoint to download.
        Default is ``None``.
    cache_dir : str or None, optional
        Path to local cache directory. If ``None`` (default), an OS-appropriate
        application-specific default cache directory is used.
    device : str or torch.device or None, optional
        Device onto which weight tensors will be mapped. If ``None``, no mapping
        is performed and tensors will be loaded onto the same device as they
        were on when saved (which will result in an error if the device is not
        present). Default is ``"cpu"``.
    return_name : bool, optional
        If ``True``, a tuple is returned indicting the name of the checkpoint
        which was loaded. This is useful if the default checkpoint was loaded.
        Default is ``False``.
    verbose : int, optional
        Verbosity level. Default is ``1``. Set to ``0`` to disable print
        statements.

    Returns
    -------
    checkpoint : dict
        Loaded checkpoint.
    checkpoint_name : str, optional
        If ``return_name`` is ``True``, the name of the checkpoint is also
        returned.
    """
    import torch

    if ckpt_name is None:
        ckpt_name = get_default_checkpoint()

    if cache_dir is None:
        cache_dir = get_default_cache_dir()

    ckpt_name_cannon = cannonise_checkpoint_name(ckpt_name)
    checkpoint_resources = get_checkpoint_list()
    builtin_ckpt_path_a = os.path.join(
        PACKAGE_DIR,
        "checkpoints",
        os.path.split(ckpt_name)[1],
    )
    builtin_ckpt_path_b = os.path.join(
        PACKAGE_DIR,
        "checkpoints",
        ckpt_name_cannon + CHECKPOINT_EXT,
    )

    using_cache = False
    if os.path.isfile(ckpt_name):
        ckpt_path = ckpt_name
        ckpt_dscr = "local"
    elif os.path.isfile(ckpt_name + CHECKPOINT_EXT):
        ckpt_path = ckpt_name + CHECKPOINT_EXT
        ckpt_dscr = "local"
    elif os.path.isfile(builtin_ckpt_path_a):
        ckpt_path = builtin_ckpt_path_a
        ckpt_dscr = "builtin"
    elif os.path.isfile(builtin_ckpt_path_b):
        ckpt_path = builtin_ckpt_path_b
        ckpt_dscr = "builtin"
    elif ckpt_name_cannon in checkpoint_resources:
        using_cache = True
        ckpt_path = download_checkpoint(
            ckpt_name_cannon, cache_dir=cache_dir, verbose=verbose
        )
        ckpt_dscr = "cached"
    else:
        msg = style.error_fmt(
            "The checkpoint parameter should either be a path to a file or one of"
        )
        msg += "\n  ".join([""] + list(checkpoint_resources.keys()))
        msg += style.error_fmt("\nbut {} was provided.".format(ckpt_name))
        with style.error_message():
            raise ValueError(msg)

    if not os.path.isfile(ckpt_path):
        msg = "No checkpoint found at '{}'".format(ckpt_path)
        with style.error_message(msg) as msg:
            raise EnvironmentError(msg)
    if verbose >= 1:
        print("Loading model from {} checkpoint:\n  '{}'".format(ckpt_dscr, ckpt_path))

    load_args = {}
    if device is not None:
        # Map model to be loaded to specified single gpu.
        load_args = dict(map_location=device)
    try:
        checkpoint = torch.load(ckpt_path, **load_args)
    except pickle.UnpicklingError:
        if not using_cache:
            # Direct path to checkpoint was given, so we shouldn't delete
            # the user's file
            msg = "Error: Unable to load checkpoint {}".format(
                os.path.abspath(ckpt_path)
            )
            with style.error_message(msg) as msg:
                print(msg)
                raise
        else:
            # Delete the checkpoint and try again, in case it is just a
            # malformed download (interrupted download, etc)
            os.remove(ckpt_path)
            ckpt_path = download_checkpoint(
                ckpt_name, cache_dir=cache_dir, verbose=verbose
            )
            try:
                checkpoint = torch.load(ckpt_path, **load_args)

            except pickle.UnpicklingError:

                msg = "Error: Unable to load checkpoint {}.".format(
                    os.path.abspath(ckpt_path)
                )
                with style.error_message(msg) as msg:
                    print(msg)

                # Check if there was an error because the file was missing
                # and we downloaded a 404 error page instead.
                with open(ckpt_path) as myfile:
                    contents = myfile.read()
                    if "Not Found" in contents or "Error 404" in contents:
                        msg = (
                            "The file you are trying to download is not available"
                            " at this web address."
                            " The download produced a 404 Error (File Not Found)."
                            "\nThe original source file may have been moved or deleted"
                            " by its host."
                        )
                        with style.error_message(msg) as msg:
                            print(msg + "\n")
                            raise EnvironmentError(msg) from None

                # Check if the user ran out of storage space part-way through
                # the download.
                import shutil

                _, _, free_B = shutil.disk_usage("/")
                free_MiB = free_B // 10**6

                if free_MiB < 64:
                    msg = (
                        "You only have {}MB of free space on your hard disk."
                        " Please free up 100MB of space on your hard disk and"
                        " then try again.\n"
                    ).format(free_MiB)
                    with style.error_message(msg) as msg:
                        print(msg + "\n")
                        raise EnvironmentError(msg) from None

                msg = "There was an unknown issue opening the downloaded file."
                with style.error_message(msg) as msg:
                    print(msg)
                    raise

    if return_name:
        return checkpoint, ckpt_name

    return checkpoint



"""
General utility functions.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import contextlib

import numpy as np
import scipy.stats


def mode(a, axis=None, keepdims=False, **kwargs):
    """
    Return an array of the modal (most common) value in the passed array.

    If there is more than one such value, only the smallest is returned.

    Parameters
    ----------
    a : array_like
        n-dimensional array of which to find mode(s).
    axis : int or None, optional
        Axis or axes along which the mode is computed. The default,
        ``axis=None``, will sum all of the elements of the input array.
        If axis is negative it counts from the last to the first axis.
    keepdims : bool, optional
        If this is set to ``True``, the axes which are reduced are left
        in the result as dimensions with size one. With this option, the result
        will broadcast correctly against the input array. Default is ``False``.
    **kwargs
        Additional arguments as per :meth:`scipy.stats.mode`.

    Returns
    -------
    mode_along_axis : numpy.ndarray
        An array with the same shape as ``a``, with the specified axis removed.
        If ``keepdims=True`` and either ``a`` is a 0-d array or ``axis`` is None,
        a scalar is returned.

    See Also
    --------
    scipy.stats.mode
    """
    m = scipy.stats.mode(a, axis=axis, **kwargs)[0]
    if keepdims:
        return m
    m = m.squeeze(0 if axis is None else axis)
    if m.size == 1:
        m = m.item()
    return m


def first_nonzero(arr, axis=-1, invalid_val=-1):
    """
    Find the index of the first non-zero element in an array.

    Parameters
    ----------
    arr : numpy.ndarray
        Array to search.
    axis : int, optional
        Axis along which to search for a non-zero element. Default is ``-1``.
    invalid_val : any, optional
        Value to return if all elements are zero. Default is ``-1``.
    """
    mask = arr != 0
    return np.where(mask.any(axis=axis), mask.argmax(axis=axis), invalid_val)


def last_nonzero(arr, axis=-1, invalid_val=-1):
    """
    Find the index of the last non-zero element in an array.

    Parameters
    ----------
    arr : numpy.ndarray
        Array to search.
    axis : int, optional
        Axis along which to search for a non-zero element. Default is ``-1``.
    invalid_val : any, optional
        Value to return if all elements are zero. Default is ``-1``.
    """
    mask = arr != 0
    val = arr.shape[axis] - np.flip(mask, axis=axis).argmax(axis=axis) - 1
    return np.where(mask.any(axis=axis), val, invalid_val)


def get_indicator_onoffsets(indicator):
    """
    Find the onsets and offsets of nonzero entries in an indicator.

    Parameters
    ----------
    indicator : 1d numpy.ndarray
        Input vector, which is sometimes zero and sometimes nonzero.

    Returns
    -------
    onsets : list
        Onset indices, where each entry is the start of a sequence of nonzero
        values in the input ``indicator``.
    offsets : list
        Offset indices, where each entry is the last in a sequence of nonzero
        values in the input ``indicator``, such that
        ``indicator[onsets[i] : offsets[i] + 1] != 0``.
    """
    indices = np.nonzero(indicator)[0]

    if len(indices) == 0:
        return [], []

    onsets = [indices[0]]
    offsets = []
    breaks = np.nonzero(indices[1:] - indices[:-1] > 1)[0]
    for break_idx in breaks:
        offsets.append(indices[break_idx])
        onsets.append(indices[break_idx + 1])
    offsets.append(indices[-1])

    return onsets, offsets
